# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
LilyPond words.

Some lists could maybe be generated dynamically.

"""

keywords = (
    'accepts',
    'alias',
    'consists',
    'defaultchild',
    'denies',
    #'description',
    #'grobdescriptions',
    'hide', # since 2.18
    'include',
    #'invalid',
    'language',
    'name',
    #'objectid',
    'omit', # since 2.18
    'once',
    'set',
    'unset',
    'override',
    'revert',
    'remove',
    'temporary', # since 2.18
    'sequential',
    'simultaneous',
    #'type',
    'undo', # since 2.18 (not mentioned in the command index)
    'version',
    'score',
    'book',
    'bookpart',
    'header',
    'paper',
    'midi',
    'layout',
    'with',
    'context',
)


markup_commands_nargs = (
# no arguments
(
    'doubleflat',
    'doublesharp',
    'eyeglasses',
    'flat',
    'natural',
    'null',
    'semiflat',
    'semisharp',
    'sesquiflat',
    'sesquisharp',
    'sharp',
    'strut',
    'table-of-contents'
),
# one argument
(
    'backslashed-digit',
    'bold',
    'box',
    'bracket',
    'caps',
    'center-align',
    'center-column',
    'char',
    'circle',
    'column',
    'concat',
    'dir-column',
    'draw-dashed-line', # since 2.18
    'draw-dotted-line', # since 2.18
    'draw-line',
    'dynamic',
    'fill-line',
    'finger',
    'fontCaps',
    'fret-diagram',
    'fret-diagram-terse',
    'fret-diagram-verbose',
    'fromproperty',
    'harp-pedal',
    'hbracket',
    'hspace',
    'huge',
    'italic',
    'justify',
    'justify-field',
    'justify-string',
    'large',
    'larger',
    'left-align',
    'left-brace',
    'left-column',
    'line',
    'lookup',
    'markalphabet',
    'markletter',
    'medium',
    'musicglyph',
    'normalsize',
    'normal-size-sub',
    'normal-size-super',
    'normal-text',
    'number',
    'oval', # since 2.18
    'postscript',
    'right-align',
    'right-brace',
    'right-column',
    'roman',
    'rounded-box',
    'sans',
    'score',
    'simple',
    'slashed-digit',
    'small',
    'smallCaps',
    'smaller',
    'stencil',
    'sub',
    'super',
    'teeny',
    'text',
    'tied-lyric',
    'tiny',
    'transparent',
    'triangle',
    'typewriter',
    'underline',
    'upright',
    'vcenter',
    'vspace',
    'verbatim-file',
    'whiteout',
    'wordwrap',
    'wordwrap-field',
    'wordwrap-string',
),
# two arguments
(
    'abs-fontsize',
    'auto-footnote', # since 2.16
    'combine',
    'customTabClef',
    'fontsize',
    'footnote',
    'fraction',
    'halign',
    'hcenter-in',
    'lower',
    'magnify',
    'note',
    'on-the-fly',
    'override',
    'pad-around',
    'pad-markup',
    'pad-x',
    'page-link',
    'path',     # added in LP 2.13.31
    'raise',
    'rotate',
    'scale',
    'translate',
    'translate-scaled',
    'with-color',
    'with-link',
    'with-url',
    'woodwind-diagram',
),
# three arguments
(
    'arrow-head',
    'beam',
    'draw-circle',
    'epsfile',
    'filled-box',
    'general-align',
    'note-by-number',
    'pad-to-box',
    'page-ref',
    'with-dimensions',
),
# four arguments
(
    'pattern',
    'put-adjacent',
),
# five arguments,
(
    'fill-with-pattern',
),
)

markup_commands = sum(markup_commands_nargs, ())


music_commands = (
    'absolute', # since 2.18
    'acciaccatura',
    'accidentalStyle', # since 2.16
    'addChordShape', # since 2.16
    'addInstrumentDefinition',
    'addlyrics',
    'addQuote',
    'afterGrace',
    #'afterGraceFraction', # this is a parser variable
    'aikenHeads',
    'aikenHeadsMinor',
    'allowPageTurn',
    'alterBroken', # since 2.18 (?)
    'alternative',
    #'AncientRemoveEmptyStaffContext',
    'appendToTag', # since 2.16
    'applyContext',
    'applyMusic',
    'applyOutput',
    'appoggiatura',
    'arpeggio',
    'arpeggioArrowDown',
    'arpeggioArrowUp',
    'arpeggioBracket',
    'arpeggioNormal',
    'arpeggioParenthesis',
    'ascendens',
    'auctum',
    'augmentum',
    'autoAccidentals',
    'autoBeamOff',
    'autoBeamOn',
    'autochange',
    'balloonGrobText',
    'balloonLengthOff',
    'balloonLengthOn',
    'balloonText',
    'bar',
    'barNumberCheck',
    'bassFigureExtendersOff',
    'bassFigureExtendersOn',
    'bassFigureStaffAlignmentDown',
    'bassFigureStaffAlignmentNeutral',
    'bassFigureStaffAlignmentUp',
    'bendAfter',
    'blackTriangleMarkup',
    'bookOutputName',
    'bookOutputSuffix',
    'bracketCloseSymbol',
    'bracketOpenSymbol',
    'break',
    'breathe',
    'breve',
    'cadenzaOff',
    'cadenzaOn',
    'caesura',
    'cavum',
    'change',
    'chordmode',
    #'chordNameSeparator',
    #'chordPrefixSpacer',
    #'chordRootNamer',
    'chordRepeats', # since 2.16
    'chords',
    'clef',
    'cm',
    'compoundMeter', # since 2.16
    'compressFullBarRests',
    'context',
    'cr',
    'cresc',
    'crescHairpin',
    'crescTextCresc',
    'crossStaff', # since 2.16
    'cueClef',  # since 2.16
    'cueClefUnset',  # since 2.16
    'cueDuring',
    'cueDuringWithClef',  # since 2.16
    'dashBar',
    'dashDash',
    'dashDot',
    'dashHat',
    'dashLarger',
    'dashPlus',
    'dashUnderscore',
    'deadNote',  # since 2.16
    'decr',
    'default',
    'defaultNoteHeads',  # since 2.16
    'defaultTimeSignature',
    'defineBarLine', # since 2.18
    'deminutum',
    'denies',
    'descendens',
    'dim',
    'dimHairpin',
    'dimTextDecr',
    'dimTextDecresc',
    'dimTextDim',
    'displayLilyMusic',
    'displayMusic',
    'divisioMaior',
    'divisioMaxima',
    'divisioMinima',
    'dotsDown',
    'dotsNeutral',
    'dotsUp',
    'drummode',
    'drumPitchTable',
    'drums',
    'dynamicDown',
    'dynamicNeutral',
    'dynamicUp',
    'easyHeadsOff',
    'easyHeadsOn',
    'endcr',
    'endcresc',
    'enddecr',
    'enddim',
    'endincipit',
    'endSpanners',
    'episemFinis',
    'episemInitium',
    'escapedBiggerSymbol',
    'escapedExclamationSymbol',
    'escapedParenthesisCloseSymbol',
    'escapedParenthesisOpenSymbol',
    'escapedSmallerSymbol',
    'expandFullBarRests',
    'f',
    'featherDurations',
    'fermataMarkup',
    'ff',
    'fff',
    'ffff',
    'fffff',
    'figuremode',
    'figures',
    'finalis',
    'fingeringOrientations',
    'flexa',
    'fp',
    'frenchChords',
    'fullJazzExceptions',
    'funkHeads',
    'funkHeadsMinor',
    'fz',
    'germanChords',
    'glissando',
    'grace',
    'graceSettings',
    'harmonic',
    'hideNotes',
    'hideStaffSwitch',
    'huge',
    'ignatzekExceptionMusic',
    'ignatzekExceptions',
    'iij',
    'IIJ',
    'ij',
    'IJ',
    'improvisationOff',
    'improvisationOn',
    'in',
    'inclinatum',
    'includePageLayoutFile',
    'indent',
    'instrumentSwitch',
    'instrumentTransposition',
    'interscoreline',
    'italianChords',
    'keepWithTag',
    'key',
    'killCues',
    'label',
    'laissezVibrer',
    'large',
    'ligature',
    'linea',
    'longa',
    'lyricmode',
    'lyrics',
    'lyricsto',
    'maininput',
    'majorSevenSymbol',
    'makeClusters',
    'mark',
    'markLengthOff', # since 2.18
    'markLengthOn',  # since 2.18
    'markup',
    'markuplines', # deprecated, till 2.14
    'markuplist', # from 2.16
    'maxima',
    'melisma',
    'melismaEnd',
    'mergeDifferentlyDottedOff',
    'mergeDifferentlyDottedOn',
    'mergeDifferentlyHeadedOff',
    'mergeDifferentlyHeadedOn',
    'mf',
    'mm',
    'mp',
    'musicMap',
    'neumeDemoLayout',
    'new',
    'newSpacingSection',
    'noBeam',
    'noBreak',
    'noPageBreak',
    'noPageTurn',
    'normalsize',
    'notemode',
    'numericTimeSignature',
    'octaveCheck',
    'oldaddlyrics',
    'oneVoice',
    'oriscus',
    'ottava',
    'override',
    'overrideProperty',
    'overrideTimeSignatureSettings',  # since 2.16
    'p',
    'pageBreak',
    'pageTurn',
    'palmMute',  # since 2.16
    'palmMuteOn',  # since 2.16
    'parallelMusic',
    'parenthesisCloseSymbol',
    'parenthesisOpenSymbol',
    'parenthesize',
    'partcombine',
    'partCombineListener',
    'partial',
    'partialJazzExceptions',
    'partialJazzMusic',
    'pes',
    'phrasingSlurDashed',
    'phrasingSlurDotted',
    'phrasingSlurDown',
    'phrasingSlurNeutral',
    'phrasingSlurSolid',
    'phrasingSlurUp',
    'pipeSymbol',
    'pitchedTrill',
    'pointAndClickOff',
    'pointAndClickOn',
    'pp',
    'ppp',
    'pppp',
    'ppppp',
    'predefinedFretboardsOff',
    'predefinedFretboardsOn',
    'pt',
    'pushToTag', # since 2.16
    'quilisma',
    'quoteDuring',
    'relative',
    'RemoveEmptyRhythmicStaffContext',
    'RemoveEmptyStaffContext',
    'removeWithTag',
    'repeat',
    'repeatTie',
    'resetRelativeOctave',
    'responsum',
    'rest',
    'revert',
    'rfz',
    'rightHandFinger',
    'sacredHarpHeads',
    'sacredHarpHeadsMinor',
    'scaleDurations',
    'scoreTweak',
    'semiGermanChords',
    'set',
    'sf',
    'sff',
    'sfp',
    'sfz',
    'shape', # since 2.16
    'shiftDurations',
    'shiftOff',
    'shiftOn',
    'shiftOnn',
    'shiftOnnn',
    'showStaffSwitch',
    'single', # since 2.18
    'skip',
    'skipTypesetting',
    'slurDashed',
    'slurDotted',
    'slurDown',
    'slurNeutral',
    'slurSolid',
    'slurUp',
    'small',
    'sostenutoOff',
    'sostenutoOn',
    'southernHarmonyHeads',
    'southernHarmonyHeadsMinor',
    'sp',
    'spacingTweaks',
    'spp',
    'startAcciaccaturaMusic',
    'startAppoggiaturaMusic',
    'startGraceMusic',
    'startGroup',
    'startStaff',
    'startTextSpan',
    'startTrillSpan',
    'stemDown',
    'stemNeutral',
    'stemUp',
    'stopAcciaccaturaMusic',
    'stopAppoggiaturaMusic',
    'stopGraceMusic',
    'stopGroup',
    'stopStaff',
    'stopTextSpan',
    'stopTrillSpan',
    'stringTuning', # since 2.16
    'strokeFingerOrientations',
    'stropha',
    'sustainOff',
    'sustainOn',
    'tabFullNotation',
    'tag',
    'teeny',
    'tempo',
    'tempoWholesPerMinute',
    'textLengthOff',
    'textLengthOn',
    'textSpannerDown',
    'textSpannerNeutral',
    'textSpannerUp',
    'tieDashed',
    'tieDotted',
    'tieDown',
    'tieNeutral',
    'tieSolid',
    'tieUp',
    'tildeSymbol',
    'time',
    'times',
    'timing',
    'tiny',
    'tocItem', # since ?
    'transpose',
    'transposedCueDuring',
    'transposition',
    'treCorde',
    'tuplet', # since 2.18
    'tupletDown',
    'tupletNeutral',
    'tupletUp',
    'tweak',
    'unaCorda',
    'unfoldRepeats',
    'unHideNotes',
    'unit',
    'unset',
    'versus',
    'virga',
    'virgula',
    'voiceFour',
    'voiceFourStyle',
    'voiceNeutralStyle',
    'voiceOne',
    'voiceOneStyle',
    'voiceThree',
    'voiceThreeStyle',
    'voiceTwo',
    'voiceTwoStyle',
    'walkerHeads',
    'walkerHeadsMinor',
    'whiteTriangleMarkup',
    'withMusicProperty',
)


articulations = (
    'accent',
    'espressivo',
    'marcato',
    'portato',
    'staccatissimo',
    'staccato',
    'tenuto',
)


ornaments = (
    'prall',
    'mordent',
    'prallmordent',
    'turn',
    'upprall',
    'downprall',
    'upmordent',
    'downmordent',
    'lineprall',
    'prallprall',
    'pralldown',
    'prallup',
    'reverseturn',
    'trill',
)


fermatas = (
    'shortfermata',
    'fermata',
    'longfermata',
    'verylongfermata',
)


instrument_scripts = (
    'upbow',
    'downbow',
    'flageolet',
    'thumb',
    'snappizzicato',
    'open',
    'halfopen',
    'stopped',
    'lheel',
    'rheel',
    'ltoe',
    'rtoe',
)


repeat_scripts = (
    'segno',
    'coda',
    'varcoda',
)


ancient_scripts = (
    'ictus',
    'accentus',
    'circulus',
    'semicirculus',
    'signumcongruentiae',
)


all_articulations = (
    articulations + ornaments + fermatas + instrument_scripts +
    repeat_scripts + ancient_scripts
)


modes = (
    'major',
    'minor',
    'ionian',
    'dorian',
    'phrygian',
    'lydian',
    'mixolydian',
    'aeolian',
    'locrian',
)


contexts = (
    'ChoirStaff',
    'ChordNames',
    'CueVoice',
    'Devnull',
    'DrumStaff',
    'DrumVoice',
    'Dynamics',
    'FiguredBass',
    'FretBoards',
    'Global',
    'GrandStaff',
    'GregorianTranscriptionStaff',
    'GregorianTranscriptionVoice',
    'KievanStaff', # since 2.16
    'KievanVoice', # since 2.16
    'Lyrics',
    'MensuralStaff',
    'MensuralVoice',
    'NoteNames',
    'NullVoice',     # since 2.18
    'PetrucciStaff', # since 2.16
    'PetrucciVoice', # since 2.16
    'PianoStaff',
    'RhythmicStaff',
    'Score',
    'Staff',
    'StaffGroup',
    'TabStaff',
    'TabVoice',
    'Timing',
    'VaticanaStaff',
    'VaticanaVoice',
    'Voice',
)

grobs = (
    'Accidental',
    'AccidentalCautionary',
    'AccidentalPlacement',
    'AccidentalSuggestion',
    'Ambitus',
    'AmbitusAccidental',
    'AmbitusLine',
    'AmbitusNoteHead',
    'Arpeggio',
    'BalloonTextItem',
    'BarLine',
    'BarNumber',
    'BassFigure',
    'BassFigureAlignment',
    'BassFigureAlignmentPositioning',
    'BassFigureBracket',
    'BassFigureContinuation',
    'BassFigureLine',
    'Beam',
    'BendAfter',
    'BreakAlignGroup',
    'BreakAlignment',
    'BreathingSign',
    'ChordName',
    'Clef',
    'ClefModifier',
    'ClusterSpanner',
    'ClusterSpannerBeacon',
    'CombineTextScript',
    'CueClef',
    'CueEndClef',
    'Custos',
    'DotColumn',
    'Dots',
    'DoublePercentRepeat',
    'DoublePercentRepeatCounter',
    'DoubleRepeatSlash',
    'DynamicLineSpanner',
    'DynamicText',
    'DynamicTextSpanner',
    'Episema',
    'Fingering',
    'FingeringColumn',
    'Flag',
    'FootnoteItem',
    'FootnoteSpanner',
    'FretBoard',
    'Glissando',
    'GraceSpacing',
    'GridLine',
    'GridPoint',
    'Hairpin',
    'HorizontalBracket',
    'InstrumentName',
    'InstrumentSwitch',
    'KeyCancellation',
    'KeySignature',
    'KievanLigature',
    'LaissezVibrerTie',
    'LaissezVibrerTieColumn',
    'LedgerLineSpanner',
    'LeftEdge',
    'LigatureBracket',
    'LyricExtender',
    'LyricHyphen',
    'LyricSpace',
    'LyricText',
    'MeasureCounter',
    'MeasureGrouping',
    'MelodyItem',
    'MensuralLigature',
    'MetronomeMark',
    'MultiMeasureRest',
    'MultiMeasureRestNumber',
    'MultiMeasureRestText',
    'NonMusicalPaperColumn',
    'NoteCollision',
    'NoteColumn',
    'NoteHead',
    'NoteName',
    'NoteSpacing',
    'OttavaBracket',
    'PaperColumn',
    'ParenthesesItem',
    'PercentRepeat',
    'PercentRepeatCounter',
    'PhrasingSlur',
    'PianoPedalBracket',
    'RehearsalMark',
    'RepeatSlash',
    'RepeatTie',
    'RepeatTieColumn',
    'Rest',
    'RestCollision',
    'Script',
    'ScriptColumn',
    'ScriptRow',
    'Slur',
    'SostenutoPedal',
    'SostenutoPedalLineSpanner',
    'SpacingSpanner',
    'SpanBar',
    'SpanBarStub',
    'StaffGrouper',
    'StaffSpacing',
    'StaffSymbol',
    'StanzaNumber',
    'Stem',
    'StemStub',
    'StemTremolo',
    'StringNumber',
    'StrokeFinger',
    'SustainPedal',
    'SustainPedalLineSpanner',
    'System',
    'SystemStartBar',
    'SystemStartBrace',
    'SystemStartBracket',
    'SystemStartSquare',
    'TabNoteHead',
    'TextScript',
    'TextSpanner',
    'Tie',
    'TieColumn',
    'TimeSignature',
    'TrillPitchAccidental',
    'TrillPitchGroup',
    'TrillPitchHead',
    'TrillSpanner',
    'TupletBracket',
    'TupletNumber',
    'UnaCordaPedal',
    'UnaCordaPedalLineSpanner',
    'VaticanaLigature',
    'VerticalAlignment',
    'VerticalAxisGroup',
    'VoiceFollower',
    'VoltaBracket',
    'VoltaBracketSpanner',
)


DOUBLE_FLAT     = -1
THREE_Q_FLAT    = -0.75
FLAT            = -0.5
SEMI_FLAT       = -0.25
NATURAL         =  0
SEMI_SHARP      =  0.25
SHARP           =  0.5
THREE_Q_SHARP   =  0.75
DOUBLE_SHARP    =  1

pitch_names = {
    "nederlands": {
        "ceses": (0, DOUBLE_FLAT),
        "ceseh": (0, THREE_Q_FLAT),
        "ces": (0, FLAT),
        "ceh": (0, SEMI_FLAT),
        "c": (0, NATURAL),
        "cih": (0, SEMI_SHARP),
        "cis": (0, SHARP),
        "cisih": (0, THREE_Q_SHARP),
        "cisis": (0, DOUBLE_SHARP),

        "deses": (1, DOUBLE_FLAT),
        "deseh": (1, THREE_Q_FLAT),
        "des": (1, FLAT),
        "deh": (1, SEMI_FLAT),
        "d": (1, NATURAL),
        "dih": (1, SEMI_SHARP),
        "dis": (1, SHARP),
        "disih": (1, THREE_Q_SHARP),
        "disis": (1, DOUBLE_SHARP),

        "eeses": (2, DOUBLE_FLAT),
        "eses": (2, DOUBLE_FLAT),
        "eeseh": (2, THREE_Q_FLAT),
        "ees": (2, FLAT),
        "es": (2, FLAT),
        "eeh": (2, SEMI_FLAT),
        "e": (2, NATURAL),
        "eih": (2, SEMI_SHARP),
        "eis": (2, SHARP),
        "eisih": (2, THREE_Q_SHARP),
        "eisis": (2, DOUBLE_SHARP),

        "feses": (3, DOUBLE_FLAT),
        "feseh": (3, THREE_Q_FLAT),
        "fes": (3, FLAT),
        "feh": (3, SEMI_FLAT),
        "f": (3, NATURAL),
        "fih": (3, SEMI_SHARP),
        "fis": (3, SHARP),
        "fisih": (3, THREE_Q_SHARP),
        "fisis": (3, DOUBLE_SHARP),

        "geses": (4, DOUBLE_FLAT),
        "geseh": (4, THREE_Q_FLAT),
        "ges": (4, FLAT),
        "geh": (4, SEMI_FLAT),
        "g": (4, NATURAL),
        "gih": (4, SEMI_SHARP),
        "gis": (4, SHARP),
        "gisih": (4, THREE_Q_SHARP),
        "gisis": (4, DOUBLE_SHARP),

        "aeses": (5, DOUBLE_FLAT),
        "ases": (5, DOUBLE_FLAT),
        "aeseh": (5, THREE_Q_FLAT),
        "aes": (5, FLAT),
        "as": (5, FLAT),
        "aeh": (5, SEMI_FLAT),
        "a": (5, NATURAL),
        "aih": (5, SEMI_SHARP),
        "ais": (5, SHARP),
        "aisih": (5, THREE_Q_SHARP),
        "aisis": (5, DOUBLE_SHARP),

        "beses": (6, DOUBLE_FLAT),
        "beseh": (6, THREE_Q_FLAT),
        "bes": (6, FLAT),
        "beh": (6, SEMI_FLAT),
        "b": (6, NATURAL),
        "bih": (6, SEMI_SHARP),
        "bis": (6, SHARP),
        "bisih": (6, THREE_Q_SHARP),
        "bisis": (6, DOUBLE_SHARP),
    },
    "catalan": {
        "dobb": (0, DOUBLE_FLAT),
        "dob": (0, FLAT),
        "do": (0, NATURAL),
        "dod": (0, SHARP),
        "dodd": (0, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "reb": (1, FLAT),
        "re": (1, NATURAL),
        "red": (1, SHARP),
        "redd": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mib": (2, FLAT),
        "mi": (2, NATURAL),
        "mid": (2, SHARP),
        "midd": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fab": (3, FLAT),
        "fa": (3, NATURAL),
        "fad": (3, SHARP),
        "fadd": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "solb": (4, FLAT),
        "sol": (4, NATURAL),
        "sold": (4, SHARP),
        "soldd": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "lab": (5, FLAT),
        "la": (5, NATURAL),
        "lad": (5, SHARP),
        "ladd": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sib": (6, FLAT),
        "si": (6, NATURAL),
        "sid": (6, SHARP),
        "sidd": (6, DOUBLE_SHARP),

        # deprecated?
        "dos": (0, SHARP),
        "doss": (0, DOUBLE_SHARP),
        "res": (1, SHARP),
        "ress": (1, DOUBLE_SHARP),
        "mis": (2, SHARP),
        "miss": (2, DOUBLE_SHARP),
        "fas": (3, SHARP),
        "fass": (3, DOUBLE_SHARP),
        "sols": (4, SHARP),
        "solss": (4, DOUBLE_SHARP),
        "las": (5, SHARP),
        "lass": (5, DOUBLE_SHARP),
        "sis": (6, SHARP),
        "siss": (6, DOUBLE_SHARP),
    },
    "deutsch": {
        "ceses": (0, DOUBLE_FLAT),
        "ceseh": (0, THREE_Q_FLAT),
        "ces": (0, FLAT),
        "ceh": (0, SEMI_FLAT),
        "c": (0, NATURAL),
        "cih": (0, SEMI_SHARP),
        "cis": (0, SHARP),
        "cisih": (0, THREE_Q_SHARP),
        "cisis": (0, DOUBLE_SHARP),

        "deses": (1, DOUBLE_FLAT),
        "deseh": (1, THREE_Q_FLAT),
        "des": (1, FLAT),
        "deh": (1, SEMI_FLAT),
        "d": (1, NATURAL),
        "dih": (1, SEMI_SHARP),
        "dis": (1, SHARP),
        "disih": (1, THREE_Q_SHARP),
        "disis": (1, DOUBLE_SHARP),

        "eses": (2, DOUBLE_FLAT),
        "eseh": (2, THREE_Q_FLAT),
        "es": (2, FLAT),
        "eeh": (2, SEMI_FLAT), # should be eh; kept for backward compatibility
        "eh": (2, SEMI_FLAT),
        "e": (2, NATURAL),
        "eih": (2, SEMI_SHARP),
        "eis": (2, SHARP),
        "eisih": (2, THREE_Q_SHARP),
        "eisis": (2, DOUBLE_SHARP),

        "feses": (3, DOUBLE_FLAT),
        "feseh": (3, THREE_Q_FLAT),
        "fes": (3, FLAT),
        "feh": (3, SEMI_FLAT),
        "f": (3, NATURAL),
        "fih": (3, SEMI_SHARP),
        "fis": (3, SHARP),
        "fisih": (3, THREE_Q_SHARP),
        "fisis": (3, DOUBLE_SHARP),

        "geses": (4, DOUBLE_FLAT),
        "geseh": (4, THREE_Q_FLAT),
        "ges": (4, FLAT),
        "geh": (4, SEMI_FLAT),
        "g": (4, NATURAL),
        "gih": (4, SEMI_SHARP),
        "gis": (4, SHARP),
        "gisih": (4, THREE_Q_SHARP),
        "gisis": (4, DOUBLE_SHARP),

        "asas": (5, DOUBLE_FLAT),
        "ases": (5, DOUBLE_FLAT),   # non-standard name for asas
        "asah": (5, THREE_Q_FLAT),
        "aseh": (5, THREE_Q_FLAT),  # non-standard name for asah
        "as": (5, FLAT),
        "aeh": (5, SEMI_FLAT),  # should be ah; kepy for backward compatibility
        "ah": (5, SEMI_FLAT),
        "a": (5, NATURAL),
        "aih": (5, SEMI_SHARP),
        "ais": (5, SHARP),
        "aisih": (5, THREE_Q_SHARP),
        "aisis": (5, DOUBLE_SHARP),

        "heses": (6, DOUBLE_FLAT),
        "heseh": (6, THREE_Q_FLAT),
        "b": (6, FLAT),
        "beh": (6, SEMI_FLAT),
        "h": (6, NATURAL),
        "hih": (6, SEMI_SHARP),
        "his": (6, SHARP),
        "hisih": (6, THREE_Q_SHARP),
        "hisis": (6, DOUBLE_SHARP),
    },
    "english": {
        "cff": (0, DOUBLE_FLAT),
        "ctqf": (0, THREE_Q_FLAT),
        "cf": (0, FLAT),
        "cqf": (0, SEMI_FLAT),
        "c": (0, NATURAL),
        "cqs": (0, SEMI_SHARP),
        "cs": (0, SHARP),
        "ctqs": (0, THREE_Q_SHARP),
        "css": (0, DOUBLE_SHARP),
        "cx": (0, DOUBLE_SHARP),

        "dff": (1, DOUBLE_FLAT),
        "dtqf": (1, THREE_Q_FLAT),
        "df": (1, FLAT),
        "dqf": (1, SEMI_FLAT),
        "d": (1, NATURAL),
        "dqs": (1, SEMI_SHARP),
        "ds": (1, SHARP),
        "dtqs": (1, THREE_Q_SHARP),
        "dss": (1, DOUBLE_SHARP),
        "dx": (1, DOUBLE_SHARP),

        "eff": (2, DOUBLE_FLAT),
        "etqf": (2, THREE_Q_FLAT),
        "ef": (2, FLAT),
        "eqf": (2, SEMI_FLAT),
        "e": (2, NATURAL),
        "eqs": (2, SEMI_SHARP),
        "es": (2, SHARP),
        "etqs": (2, THREE_Q_SHARP),
        "ess": (2, DOUBLE_SHARP),
        "ex": (2, DOUBLE_SHARP),

        "fff": (3, DOUBLE_FLAT),
        "ftqf": (3, THREE_Q_FLAT),
        "ff": (3, FLAT),
        "fqf": (3, SEMI_FLAT),
        "f": (3, NATURAL),
        "fqs": (3, SEMI_SHARP),
        "fs": (3, SHARP),
        "ftqs": (3, THREE_Q_SHARP),
        "fss": (3, DOUBLE_SHARP),
        "fx": (3, DOUBLE_SHARP),

        "gff": (4, DOUBLE_FLAT),
        "gtqf": (4, THREE_Q_FLAT),
        "gf": (4, FLAT),
        "gqf": (4, SEMI_FLAT),
        "g": (4, NATURAL),
        "gqs": (4, SEMI_SHARP),
        "gs": (4, SHARP),
        "gtqs": (4, THREE_Q_SHARP),
        "gss": (4, DOUBLE_SHARP),
        "gx": (4, DOUBLE_SHARP),

        "aff": (5, DOUBLE_FLAT),
        "atqf": (5, THREE_Q_FLAT),
        "af": (5, FLAT),
        "aqf": (5, SEMI_FLAT),
        "a": (5, NATURAL),
        "aqs": (5, SEMI_SHARP),
        "as": (5, SHARP),
        "atqs": (5, THREE_Q_SHARP),
        "ass": (5, DOUBLE_SHARP),
        "ax": (5, DOUBLE_SHARP),

        "bff": (6, DOUBLE_FLAT),
        "btqf": (6, THREE_Q_FLAT),
        "bf": (6, FLAT),
        "bqf": (6, SEMI_FLAT),
        "b": (6, NATURAL),
        "bqs": (6, SEMI_SHARP),
        "bs": (6, SHARP),
        "btqs": (6, THREE_Q_SHARP),
        "bss": (6, DOUBLE_SHARP),
        "bx": (6, DOUBLE_SHARP),

        "c-flatflat": (0, DOUBLE_FLAT),
        "c-flat": (0, FLAT),
        "c-natural": (0, NATURAL),
        "c-sharp": (0, SHARP),
        "c-sharpsharp": (0, DOUBLE_SHARP),

        "d-flatflat": (1, DOUBLE_FLAT),
        "d-flat": (1, FLAT),
        "d-natural": (1, NATURAL),
        "d-sharp": (1, SHARP),
        "d-sharpsharp": (1, DOUBLE_SHARP),

        "e-flatflat": (2, DOUBLE_FLAT),
        "e-flat": (2, FLAT),
        "e-natural": (2, NATURAL),
        "e-sharp": (2, SHARP),
        "e-sharpsharp": (2, DOUBLE_SHARP),

        "f-flatflat": (3, DOUBLE_FLAT),
        "f-flat": (3, FLAT),
        "f-natural": (3, NATURAL),
        "f-sharp": (3, SHARP),
        "f-sharpsharp": (3, DOUBLE_SHARP),

        "g-flatflat": (4, DOUBLE_FLAT),
        "g-flat": (4, FLAT),
        "g-natural": (4, NATURAL),
        "g-sharp": (4, SHARP),
        "g-sharpsharp": (4, DOUBLE_SHARP),

        "a-flatflat": (5, DOUBLE_FLAT),
        "a-flat": (5, FLAT),
        "a-natural": (5, NATURAL),
        "a-sharp": (5, SHARP),
        "a-sharpsharp": (5, DOUBLE_SHARP),

        "b-flatflat": (6, DOUBLE_FLAT),
        "b-flat": (6, FLAT),
        "b-natural": (6, NATURAL),
        "b-sharp": (6, SHARP),
        "b-sharpsharp": (6, DOUBLE_SHARP),
    },
    "espanol": {
        "dobb": (0, DOUBLE_FLAT),
        "dotcb": (0, THREE_Q_FLAT),
        "dob": (0, FLAT),
        "docb": (0, SEMI_FLAT),
        "do": (0, NATURAL),
        "docs": (0, SEMI_SHARP),
        "dos": (0, SHARP),
        "dotcs": (0, THREE_Q_SHARP),
        "doss": (0, DOUBLE_SHARP),
        "dox": (0, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "retcb": (1, THREE_Q_FLAT),
        "reb": (1, FLAT),
        "recb": (1, SEMI_FLAT),
        "re": (1, NATURAL),
        "recs": (1, SEMI_SHARP),
        "res": (1, SHARP),
        "retcs": (1, THREE_Q_SHARP),
        "ress": (1, DOUBLE_SHARP),
        "rex": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mitcb": (2, THREE_Q_FLAT),
        "mib": (2, FLAT),
        "micb": (2, SEMI_FLAT),
        "mi": (2, NATURAL),
        "mics": (2, SEMI_SHARP),
        "mis": (2, SHARP),
        "mitcs": (2, THREE_Q_SHARP),
        "miss": (2, DOUBLE_SHARP),
        "mix": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fatcb": (3, THREE_Q_FLAT),
        "fab": (3, FLAT),
        "facb": (3, SEMI_FLAT),
        "fa": (3, NATURAL),
        "facs": (3, SEMI_SHARP),
        "fas": (3, SHARP),
        "fatcs": (3, THREE_Q_SHARP),
        "fass": (3, DOUBLE_SHARP),
        "fax": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "soltcb": (4, THREE_Q_FLAT),
        "solb": (4, FLAT),
        "solcb": (4, SEMI_FLAT),
        "sol": (4, NATURAL),
        "solcs": (4, SEMI_SHARP),
        "sols": (4, SHARP),
        "soltcs": (4, THREE_Q_SHARP),
        "solss": (4, DOUBLE_SHARP),
        "solx": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "latcb": (5, THREE_Q_FLAT),
        "lab": (5, FLAT),
        "lacb": (5, SEMI_FLAT),
        "la": (5, NATURAL),
        "lacs": (5, SEMI_SHARP),
        "las": (5, SHARP),
        "latcs": (5, THREE_Q_SHARP),
        "lass": (5, DOUBLE_SHARP),
        "lax": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sitcb": (6, THREE_Q_FLAT),
        "sib": (6, FLAT),
        "sicb": (6, SEMI_FLAT),
        "si": (6, NATURAL),
        "sics": (6, SEMI_SHARP),
        "sis": (6, SHARP),
        "sitcs": (6, THREE_Q_SHARP),
        "siss": (6, DOUBLE_SHARP),
        "six": (6, DOUBLE_SHARP),
    },
    "français": {
        "dobb": (0, DOUBLE_FLAT),
        "dobsb": (0, THREE_Q_FLAT),
        "dob": (0, FLAT),
        "dosb": (0, SEMI_FLAT),
        "do": (0, NATURAL),
        "dosd": (0, SEMI_SHARP),
        "dod": (0, SHARP),
        "dodsd": (0, THREE_Q_SHARP),
        "dodd": (0, DOUBLE_SHARP),
        "dox": (0, DOUBLE_SHARP),

        "rébb": (1, DOUBLE_FLAT),
        "rébsb": (1, THREE_Q_FLAT),
        "réb": (1, FLAT),
        "résb": (1, SEMI_FLAT),
        "ré": (1, NATURAL),
        "résd": (1, SEMI_SHARP),
        "réd": (1, SHARP),
        "rédsd": (1, THREE_Q_SHARP),
        "rédd": (1, DOUBLE_SHARP),
        "réx": (1, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "rebsb": (1, THREE_Q_FLAT),
        "reb": (1, FLAT),
        "resb": (1, SEMI_FLAT),
        "re": (1, NATURAL),
        "resd": (1, SEMI_SHARP),
        "red": (1, SHARP),
        "redsd": (1, THREE_Q_SHARP),
        "redd": (1, DOUBLE_SHARP),
        "rex": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mibsb": (2, THREE_Q_FLAT),
        "mib": (2, FLAT),
        "misb": (2, SEMI_FLAT),
        "mi": (2, NATURAL),
        "misd": (2, SEMI_SHARP),
        "mid": (2, SHARP),
        "midsd": (2, THREE_Q_SHARP),
        "midd": (2, DOUBLE_SHARP),
        "mix": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fabsb": (3, THREE_Q_FLAT),
        "fab": (3, FLAT),
        "fasb": (3, SEMI_FLAT),
        "fa": (3, NATURAL),
        "fasd": (3, SEMI_SHARP),
        "fad": (3, SHARP),
        "fadsd": (3, THREE_Q_SHARP),
        "fadd": (3, DOUBLE_SHARP),
        "fax": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "solbsb": (4, THREE_Q_FLAT),
        "solb": (4, FLAT),
        "solsb": (4, SEMI_FLAT),
        "sol": (4, NATURAL),
        "solsd": (4, SEMI_SHARP),
        "sold": (4, SHARP),
        "soldsd": (4, THREE_Q_SHARP),
        "soldd": (4, DOUBLE_SHARP),
        "solx": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "labsb": (5, THREE_Q_FLAT),
        "lab": (5, FLAT),
        "lasb": (5, SEMI_FLAT),
        "la": (5, NATURAL),
        "lasd": (5, SEMI_SHARP),
        "lad": (5, SHARP),
        "ladsd": (5, THREE_Q_SHARP),
        "ladd": (5, DOUBLE_SHARP),
        "lax": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sibsb": (6, THREE_Q_FLAT),
        "sib": (6, FLAT),
        "sisb": (6, SEMI_FLAT),
        "si": (6, NATURAL),
        "sisd": (6, SEMI_SHARP),
        "sid": (6, SHARP),
        "sidsd": (6, THREE_Q_SHARP),
        "sidd": (6, DOUBLE_SHARP),
        "six": (6, DOUBLE_SHARP),
    },
    "italiano": {
        "dobb": (0, DOUBLE_FLAT),
        "dobsb": (0, THREE_Q_FLAT),
        "dob": (0, FLAT),
        "dosb": (0, SEMI_FLAT),
        "do": (0, NATURAL),
        "dosd": (0, SEMI_SHARP),
        "dod": (0, SHARP),
        "dodsd": (0, THREE_Q_SHARP),
        "dodd": (0, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "rebsb": (1, THREE_Q_FLAT),
        "reb": (1, FLAT),
        "resb": (1, SEMI_FLAT),
        "re": (1, NATURAL),
        "resd": (1, SEMI_SHARP),
        "red": (1, SHARP),
        "redsd": (1, THREE_Q_SHARP),
        "redd": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mibsb": (2, THREE_Q_FLAT),
        "mib": (2, FLAT),
        "misb": (2, SEMI_FLAT),
        "mi": (2, NATURAL),
        "misd": (2, SEMI_SHARP),
        "mid": (2, SHARP),
        "midsd": (2, THREE_Q_SHARP),
        "midd": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fabsb": (3, THREE_Q_FLAT),
        "fab": (3, FLAT),
        "fasb": (3, SEMI_FLAT),
        "fa": (3, NATURAL),
        "fasd": (3, SEMI_SHARP),
        "fad": (3, SHARP),
        "fadsd": (3, THREE_Q_SHARP),
        "fadd": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "solbsb": (4, THREE_Q_FLAT),
        "solb": (4, FLAT),
        "solsb": (4, SEMI_FLAT),
        "sol": (4, NATURAL),
        "solsd": (4, SEMI_SHARP),
        "sold": (4, SHARP),
        "soldsd": (4, THREE_Q_SHARP),
        "soldd": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "labsb": (5, THREE_Q_FLAT),
        "lab": (5, FLAT),
        "lasb": (5, SEMI_FLAT),
        "la": (5, NATURAL),
        "lasd": (5, SEMI_SHARP),
        "lad": (5, SHARP),
        "ladsd": (5, THREE_Q_SHARP),
        "ladd": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sibsb": (6, THREE_Q_FLAT),
        "sib": (6, FLAT),
        "sisb": (6, SEMI_FLAT),
        "si": (6, NATURAL),
        "sisd": (6, SEMI_SHARP),
        "sid": (6, SHARP),
        "sidsd": (6, THREE_Q_SHARP),
        "sidd": (6, DOUBLE_SHARP),
    },
    "norsk": {
        "ceses": (0, DOUBLE_FLAT),
        "cessess": (0, DOUBLE_FLAT),
        "ces": (0, FLAT),
        "cess": (0, FLAT),
        "c": (0, NATURAL),
        "cis": (0, SHARP),
        "ciss": (0, SHARP),
        "cisis": (0, DOUBLE_SHARP),
        "cississ": (0, DOUBLE_SHARP),

        "deses": (1, DOUBLE_FLAT),
        "dessess": (1, DOUBLE_FLAT),
        "des": (1, FLAT),
        "dess": (1, FLAT),
        "d": (1, NATURAL),
        "dis": (1, SHARP),
        "diss": (1, SHARP),
        "disis": (1, DOUBLE_SHARP),
        "dississ": (1, DOUBLE_SHARP),

        "eeses": (2, DOUBLE_FLAT),
        "eessess": (2, DOUBLE_FLAT),
        "eses": (2, DOUBLE_FLAT),
        "essess": (2, DOUBLE_FLAT),
        "ees": (2, FLAT),
        "eess": (2, FLAT),
        "es": (2, FLAT),
        "ess": (2, FLAT),
        "e": (2, NATURAL),
        "eis": (2, SHARP),
        "eiss": (2, SHARP),
        "eisis": (2, DOUBLE_SHARP),
        "eississ": (2, DOUBLE_SHARP),

        "feses": (3, DOUBLE_FLAT),
        "fessess": (3, DOUBLE_FLAT),
        "fes": (3, FLAT),
        "fess": (3, FLAT),
        "f": (3, NATURAL),
        "fis": (3, SHARP),
        "fiss": (3, SHARP),
        "fisis": (3, DOUBLE_SHARP),
        "fississ": (3, DOUBLE_SHARP),

        "geses": (4, DOUBLE_FLAT),
        "gessess": (4, DOUBLE_FLAT),
        "ges": (4, FLAT),
        "gess": (4, FLAT),
        "g": (4, NATURAL),
        "g": (4, NATURAL),
        "gis": (4, SHARP),
        "giss": (4, SHARP),
        "gisis": (4, DOUBLE_SHARP),
        "gississ": (4, DOUBLE_SHARP),

        "aeses": (5, DOUBLE_FLAT),
        "aessess": (5, DOUBLE_FLAT),
        "ases": (5, DOUBLE_FLAT),
        "assess": (5, DOUBLE_FLAT),
        "aes": (5, FLAT),
        "aess": (5, FLAT),
        "as": (5, FLAT),
        "ass": (5, FLAT),
        "a": (5, NATURAL),
        "ais": (5, SHARP),
        "aiss": (5, SHARP),
        "aisis": (5, DOUBLE_SHARP),
        "aississ": (5, DOUBLE_SHARP),

        "bes": (6, DOUBLE_FLAT),
        "bess": (6, DOUBLE_FLAT),
        "b": (6, FLAT),
        "b": (6, FLAT),
        "h": (6, NATURAL),
        "his": (6, SHARP),
        "hiss": (6, SHARP),
        "hisis": (6, DOUBLE_SHARP),
        "hississ": (6, DOUBLE_SHARP),
    },
    "portugues": {
        "dobb": (0, DOUBLE_FLAT),
        "dobtqt": (0, THREE_Q_FLAT),
        "dob": (0, FLAT),
        "dobqt": (0, SEMI_FLAT),
        "do": (0, NATURAL),
        "dosqt": (0, SEMI_SHARP),
        "dos": (0, SHARP),
        "dostqt": (0, THREE_Q_SHARP),
        "doss": (0, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "rebtqt": (1, THREE_Q_FLAT),
        "reb": (1, FLAT),
        "rebqt": (1, SEMI_FLAT),
        "re": (1, NATURAL),
        "resqt": (1, SEMI_SHARP),
        "res": (1, SHARP),
        "restqt": (1, THREE_Q_SHARP),
        "ress": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mibtqt": (2, THREE_Q_FLAT),
        "mib": (2, FLAT),
        "mibqt": (2, SEMI_FLAT),
        "mi": (2, NATURAL),
        "misqt": (2, SEMI_SHARP),
        "mis": (2, SHARP),
        "mistqt": (2, THREE_Q_SHARP),
        "miss": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fabtqt": (3, THREE_Q_FLAT),
        "fab": (3, FLAT),
        "fabqt": (3, SEMI_FLAT),
        "fa": (3, NATURAL),
        "fasqt": (3, SEMI_SHARP),
        "fas": (3, SHARP),
        "fastqt": (3, THREE_Q_SHARP),
        "fass": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "solbtqt": (4, THREE_Q_FLAT),
        "solb": (4, FLAT),
        "solbqt": (4, SEMI_FLAT),
        "sol": (4, NATURAL),
        "solsqt": (4, SEMI_SHARP),
        "sols": (4, SHARP),
        "solstqt": (4, THREE_Q_SHARP),
        "solss": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "labtqt": (5, THREE_Q_FLAT),
        "lab": (5, FLAT),
        "labqt": (5, SEMI_FLAT),
        "la": (5, NATURAL),
        "lasqt": (5, SEMI_SHARP),
        "las": (5, SHARP),
        "lastqt": (5, THREE_Q_SHARP),
        "lass": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sibtqt": (6, THREE_Q_FLAT),
        "sib": (6, FLAT),
        "sibqt": (6, SEMI_FLAT),
        "si": (6, NATURAL),
        "sisqt": (6, SEMI_SHARP),
        "sis": (6, SHARP),
        "sistqt": (6, THREE_Q_SHARP),
        "siss": (6, DOUBLE_SHARP),
    },
    "suomi": {
        "ceses": (0, DOUBLE_FLAT),
        "ces": (0, FLAT),
        "c": (0, NATURAL),
        "cis": (0, SHARP),
        "cisis": (0, DOUBLE_SHARP),

        "deses": (1, DOUBLE_FLAT),
        "des": (1, FLAT),
        "d": (1, NATURAL),
        "dis": (1, SHARP),
        "disis": (1, DOUBLE_SHARP),

        "eses": (2, DOUBLE_FLAT),
        "es": (2, FLAT),
        "e": (2, NATURAL),
        "eis": (2, SHARP),
        "eisis": (2, DOUBLE_SHARP),

        "feses": (3, DOUBLE_FLAT),
        "fes": (3, FLAT),
        "f": (3, NATURAL),
        "fis": (3, SHARP),
        "fisis": (3, DOUBLE_SHARP),

        "geses": (4, DOUBLE_FLAT),
        "ges": (4, FLAT),
        "g": (4, NATURAL),
        "gis": (4, SHARP),
        "gisis": (4, DOUBLE_SHARP),

        "asas": (5, DOUBLE_FLAT),
        "ases": (5, DOUBLE_FLAT),   # non-standard name for asas
        "as": (5, FLAT),
        "a": (5, NATURAL),
        "ais": (5, SHARP),
        "aisis": (5, DOUBLE_SHARP),

        "bb": (6, DOUBLE_FLAT),    # should be bes; kept for backward compatibility
        "bes": (6, DOUBLE_FLAT),
        "heses": (6, DOUBLE_FLAT),  # non-standard name for bb
        "b": (6, FLAT),
        "h": (6, NATURAL),
        "his": (6, SHARP),
        "hisis": (6, DOUBLE_SHARP),
    },
    "svenska": {
        "cessess": (0, DOUBLE_FLAT),
        "cess": (0, FLAT),
        "c": (0, NATURAL),
        "ciss": (0, SHARP),
        "cississ": (0, DOUBLE_SHARP),

        "dessess": (1, DOUBLE_FLAT),
        "dess": (1, FLAT),
        "d": (1, NATURAL),
        "diss": (1, SHARP),
        "dississ": (1, DOUBLE_SHARP),

        "essess": (2, DOUBLE_FLAT),
        "ess": (2, FLAT),
        "e": (2, NATURAL),
        "eiss": (2, SHARP),
        "eississ": (2, DOUBLE_SHARP),

        "fessess": (3, DOUBLE_FLAT),
        "fess": (3, FLAT),
        "f": (3, NATURAL),
        "fiss": (3, SHARP),
        "fississ": (3, DOUBLE_SHARP),

        "gessess": (4, DOUBLE_FLAT),
        "gess": (4, FLAT),
        "g": (4, NATURAL),
        "giss": (4, SHARP),
        "gississ": (4, DOUBLE_SHARP),

        "assess": (5, DOUBLE_FLAT),
        "ass": (5, FLAT),
        "a": (5, NATURAL),
        "aiss": (5, SHARP),
        "aississ": (5, DOUBLE_SHARP),

        "hessess": (6, DOUBLE_FLAT),
        "b": (6, FLAT),
        "h": (6, NATURAL),
        "hiss": (6, SHARP),
        "hississ": (6, DOUBLE_SHARP),
    },
    "vlaams": {
        "dobb": (0, DOUBLE_FLAT),
        "dob": (0, FLAT),
        "do": (0, NATURAL),
        "dok": (0, SHARP),
        "dokk": (0, DOUBLE_SHARP),

        "rebb": (1, DOUBLE_FLAT),
        "reb": (1, FLAT),
        "re": (1, NATURAL),
        "rek": (1, SHARP),
        "rekk": (1, DOUBLE_SHARP),

        "mibb": (2, DOUBLE_FLAT),
        "mib": (2, FLAT),
        "mi": (2, NATURAL),
        "mik": (2, SHARP),
        "mikk": (2, DOUBLE_SHARP),

        "fabb": (3, DOUBLE_FLAT),
        "fab": (3, FLAT),
        "fa": (3, NATURAL),
        "fak": (3, SHARP),
        "fakk": (3, DOUBLE_SHARP),

        "solbb": (4, DOUBLE_FLAT),
        "solb": (4, FLAT),
        "sol": (4, NATURAL),
        "solk": (4, SHARP),
        "solkk": (4, DOUBLE_SHARP),

        "labb": (5, DOUBLE_FLAT),
        "lab": (5, FLAT),
        "la": (5, NATURAL),
        "lak": (5, SHARP),
        "lakk": (5, DOUBLE_SHARP),

        "sibb": (6, DOUBLE_FLAT),
        "sib": (6, FLAT),
        "si": (6, NATURAL),
        "sik": (6, SHARP),
        "sikk": (6, DOUBLE_SHARP),
    },
}
pitch_names["español"] = pitch_names["espanol"]
#pitch_names["francais"] = pitch_names["français"]

# a set with all the pitch names for fast membership testing
all_pitch_names = frozenset(name for d in pitch_names.values() for name in d)


drum_pitch_names = (
    'acousticbassdrum',
    'bassdrum',
    'hisidestick',
    'sidestick',
    'losidestick',
    'acousticsnare',
    'snare',
    'handclap',
    'electricsnare',
    'lowfloortom',
    'closedhihat',
    'hihat',
    'highfloortom',
    'pedalhihat',
    'lowtom',
    'openhihat',
    'halfopenhihat',
    'lowmidtom',
    'himidtom',
    'crashcymbala',
    'crashcymbal',
    'hightom',
    'ridecymbala',
    'ridecymbal',
    'chinesecymbal',
    'ridebell',
    'tambourine',
    'splashcymbal',
    'cowbell',
    'crashcymbalb',
    'vibraslap',
    'ridecymbalb',
    'mutehibongo',
    'hibongo',
    'openhibongo',
    'mutelobongo',
    'lobongo',
    'openlobongo',
    'mutehiconga',
    'muteloconga',
    'openhiconga',
    'hiconga',
    'openloconga',
    'loconga',
    'hitimbale',
    'lotimbale',
    'hiagogo',
    'loagogo',
    'cabasa',
    'maracas',
    'shortwhistle',
    'longwhistle',
    'shortguiro',
    'longguiro',
    'guiro',
    'claves',
    'hiwoodblock',
    'lowoodblock',
    'mutecuica',
    'opencuica',
    'mutetriangle',
    'triangle',
    'opentriangle',
    'oneup',
    'twoup',
    'threeup',
    'fourup',
    'fiveup',
    'onedown',
    'twodown',
    'threedown',
    'fourdown',
    'fivedown',
    'bda',
    'bd',
    'ssh',
    'ss',
    'ssl',
    'sna',
    'sn',
    'hc',
    'sne',
    'tomfl',
    'hhc',
    'hh',
    'tomfh',
    'hhp',
    'toml',
    'hho',
    'hhho',
    'tomml',
    'tommh',
    'cymca',
    'cymc',
    'tomh',
    'cymra',
    'cymr',
    'cymch',
    'rb',
    'tamb',
    'cyms',
    'cb',
    'cymcb',
    'vibs',
    'cymrb',
    'bohm',
    'boh',
    'boho',
    'bolm',
    'bol',
    'bolo',
    'cghm',
    'cglm',
    'cgho',
    'cgh',
    'cglo',
    'cgl',
    'timh',
    'timl',
    'agh',
    'agl',
    'cab',
    'mar',
    'whs',
    'whl',
    'guis',
    'guil',
    'gui',
    'cl',
    'wbh',
    'wbl',
    'cuim',
    'cuio',
    'trim',
    'tri',
    'trio',
    'tt',
    'ua',
    'ub',
    'uc',
    'ud',
    'ue',
    'da',
    'db',
    'dc',
    'dd',
    'de',
)
drum_pitch_names_set = frozenset(drum_pitch_names)




# Calculator parce example

from parce import Language, lexicon, default_target, skip
from parce.action import Number, Operator, Delimiter
from parce.transform import Transform

skip_whitespace = (r'\s+', skip)


class Calculator(Language):
    @lexicon
    def root(cls):
        yield r'\d+', Number
        yield r'\-', Operator, cls.subtract
        yield r'\+', Operator, cls.add
        yield r'\*', Operator, cls.multiply
        yield r'/', Operator, cls.divide
        yield r'\(', Delimiter, cls.parens

    @lexicon
    def parens(cls):
        yield r'\)', Delimiter, -1
        yield from cls.root

    @lexicon
    def subtract(cls):
        yield r'\d+', Number
        yield r'\*', Operator, cls.multiply
        yield r'/', Operator, cls.divide
        yield r'\(', Delimiter, cls.parens
        yield skip_whitespace
        yield default_target, -1

    @lexicon
    def add(cls):
        yield from cls.subtract

    @lexicon
    def multiply(cls):
        yield r'\d+', Number
        yield r'\(', Delimiter, cls.parens
        yield skip_whitespace
        yield default_target, -1

    @lexicon
    def divide(cls):
        yield from cls.multiply


class CalculatorTransform(Transform):
    def root(self, items):
        result = 0
        for i in items:
            if i.is_token:
                if i.action is Number:
                    result = int(i.text)
            elif i.name == "add":
                result += i.obj
            elif i.name == "subtract":
                result -= i.obj
            elif i.name == "multiply":
                result *= i.obj
            elif i.name == "divide":
                result /= i.obj
            else:   # i.name == "parens":
                result = i.obj
        return result

    parens = add = subtract = multiply = divide = root




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


CSS1_NAMED_COLORS = {
    # color name  R    G    B       # CGA name      (num)
    "black": (0, 0, 0),             # black         (00)
    "navy": (0, 0, 128),            # low blue      (01)
    "green": (0, 128, 0),           # low green     (02)
    "teal": (0, 128, 128),          # low cyan      (03)
    "maroon": (128, 0, 0),          # low red       (04)
    "purple": (128, 0, 128),        # low magenta   (05)
    "olive": (128, 128, 0),         # brown         (06)
    "silver": (192, 192, 192),      # light gray    (07)
    "gray": (128, 128, 128),        # dark gray     (08)
    "grey": (128, 128, 128),        # dark gray     (08)
    "blue": (0, 0, 255),            # high blue     (09)
    "lime": (0, 255, 0),            # high green    (10)
    "aqua": (0, 255, 255),          # high cyan     (11)
    "red": (255, 0, 0),             # high red      (12)
    "fuchsia": (255, 0, 255),       # high magenta  (13)
    "yellow": (255, 255, 0),        # yellow        (14)
    "white": (255, 255, 255),       # white         (15)
}


CSS3_NAMED_COLORS = dict(**CSS1_NAMED_COLORS, **{
    # color name  R    G    B
    "aliceblue": (240, 248, 255),
    "antiquewhite": (250, 235, 215),
    "aquamarine": (127, 255, 212),
    "azure": (240, 255, 255),
    "beige": (245, 245, 220),
    "bisque": (255, 228, 196),
    "blanchedalmond": (255, 235, 205),
    "blueviolet": (138, 43, 226),
    "brown": (165, 42, 42),
    "burlywood": (222, 184, 135),
    "cadetblue": (95, 158, 160),
    "chartreuse": (127, 255, 0),
    "chocolate": (210, 105, 30),
    "coral": (255, 127, 80),
    "cornflowerblue": (100, 149, 237),
    "cornsilk": (255, 248, 220),
    "crimson": (220, 20, 60),
    "cyan": (0, 255, 255),
    "darkblue": (0, 0, 139),
    "darkcyan": (0, 139, 139),
    "darkgoldenrod": (184, 134, 11),
    "darkgray": (169, 169, 169),
    "darkgrey": (169, 169, 169),
    "darkgreen": (0, 100, 0),
    "darkkhaki": (189, 183, 107),
    "darkmagenta": (139, 0, 139),
    "darkolivegreen": (85, 107, 47),
    "darkorange": (255, 140, 0),
    "darkorchid": (153, 50, 204),
    "darkred": (139, 0, 0),
    "darksalmon": (233, 150, 122),
    "darkseagreen": (143, 188, 143),
    "darkslateblue": (72, 61, 139),
    "darkslategray": (47, 79, 79),
    "darkslategrey": (47, 79, 79),
    "darkturquoise": (0, 206, 209),
    "darkviolet": (148, 0, 211),
    "deeppink": (255, 20, 147),
    "deepskyblue": (0, 191, 255),
    "dimgray": (105, 105, 105),
    "dimgrey": (105, 105, 105),
    "dodgerblue": (30, 144, 255),
    "firebrick": (178, 34, 34),
    "floralwhite": (255, 250, 240),
    "forestgreen": (34, 139, 34),
    "gainsboro": (220, 220, 220),
    "ghostwhite": (248, 248, 255),
    "gold": (255, 215, 0),
    "goldenrod": (218, 165, 32),
    "greenyellow": (173, 255, 47),
    "honeydew": (240, 255, 240),
    "hotpink": (255, 105, 180),
    "indianred": (205, 92, 92),
    "indigo": (75, 0, 130),
    "ivory": (255, 255, 240),
    "khaki": (240, 230, 140),
    "lavender": (230, 230, 250),
    "lavenderblush": (255, 240, 245),
    "lawngreen": (124, 252, 0),
    "lemonchiffon": (255, 250, 205),
    "lightblue": (173, 216, 230),
    "lightcoral": (240, 128, 128),
    "lightcyan": (224, 255, 255),
    "lightgoldenrodyellow": (250, 250, 210),
    "lightgray": (211, 211, 211),
    "lightgrey": (211, 211, 211),
    "lightgreen": (144, 238, 144),
    "lightpink": (255, 182, 193),
    "lightsalmon": (255, 160, 122),
    "lightseagreen": (32, 178, 170),
    "lightskyblue": (135, 206, 250),
    "lightslategray": (119, 136, 153),
    "lightslategrey": (119, 136, 153),
    "lightsteelblue": (176, 196, 222),
    "lightyellow": (255, 255, 224),
    "limegreen": (50, 205, 50),
    "linen": (250, 240, 230),
    "magenta": (255, 0, 255),
    "mediumaquamarine": (102, 205, 170),
    "mediumblue": (0, 0, 205),
    "mediumorchid": (186, 85, 211),
    "mediumpurple": (147, 112, 219),
    "mediumseagreen": (60, 179, 113),
    "mediumslateblue": (123, 104, 238),
    "mediumspringgreen": (0, 250, 154),
    "mediumturquoise": (72, 209, 204),
    "mediumvioletred": (199, 21, 133),
    "midnightblue": (25, 25, 112),
    "mintcream": (245, 255, 250),
    "mistyrose": (255, 228, 225),
    "moccasin": (255, 228, 181),
    "navajowhite": (255, 222, 173),
    "oldlace": (253, 245, 230),
    "olivedrab": (107, 142, 35),
    "orange": (255, 165, 0),
    "orangered": (255, 69, 0),
    "orchid": (218, 112, 214),
    "palegoldenrod": (238, 232, 170),
    "palegreen": (152, 251, 152),
    "paleturquoise": (175, 238, 238),
    "palevioletred": (219, 112, 147),
    "papayawhip": (255, 239, 213),
    "peachpuff": (255, 218, 185),
    "peru": (205, 133, 63),
    "pink": (255, 192, 203),
    "plum": (221, 160, 221),
    "powderblue": (176, 224, 230),
    "rosybrown": (188, 143, 143),
    "royalblue": (65, 105, 225),
    "saddlebrown": (139, 69, 19),
    "salmon": (250, 128, 114),
    "sandybrown": (244, 164, 96),
    "seagreen": (46, 139, 87),
    "seashell": (255, 245, 238),
    "sienna": (160, 82, 45),
    "skyblue": (135, 206, 235),
    "slateblue": (106, 90, 205),
    "slategray": (112, 128, 144),
    "slategrey": (112, 128, 144),
    "snow": (255, 250, 250),
    "springgreen": (0, 255, 127),
    "steelblue": (70, 130, 180),
    "tan": (210, 180, 140),
    "thistle": (216, 191, 216),
    "tomato": (255, 99, 71),
    "turquoise": (64, 224, 208),
    "violet": (238, 130, 238),
    "wheat": (245, 222, 179),
    "whitesmoke": (245, 245, 245),
    "yellowgreen": (154, 205, 50),
})


# source: https://www.w3.org/Style/CSS/all-properties
CSS3_ALL_PROPERTIES = (
    "align-content",
    "align-items",
    "alignment-baseline",
    "align-self",
    "all",
    "animation",
    "animation-delay",
    "animation-direction",
    "animation-duration",
    "animation-fill-mode",
    "animation-iteration-count",
    "animation-name",
    "animation-play-state",
    "animation-timing-function",
    "appearance",
    "azimuth",
    "backface-visibility",
    "background",
    "background-attachment",
    "background-blend-mode",
    "background-clip",
    "background-color",
    "background-image",
    "background-origin",
    "background-position",
    "background-repeat",
    "background-size",
    "baseline-shift",
    "baseline-source",
    "block-ellipsis",
    "block-overflow",
    "block-size",
    "block-step",
    "block-step-align",
    "block-step-insert",
    "block-step-round",
    "block-step-size",
    "bookmark-label",
    "bookmark-level",
    "bookmark-state",
    "border",
    "border-block",
    "border-block-color",
    "border-block-end",
    "border-block-end-color",
    "border-block-end-style",
    "border-block-end-width",
    "border-block-start",
    "border-block-start-color",
    "border-block-start-style",
    "border-block-start-width",
    "border-block-style",
    "border-block-width",
    "border-bottom",
    "border-bottom-color",
    "border-bottom-left-radius",
    "border-bottom-right-radius",
    "border-bottom-style",
    "border-bottom-width",
    "border-boundary",
    "border-collapse",
    "border-color",
    "border-end-end-radius",
    "border-end-start-radius",
    "border-image",
    "border-image-outset",
    "border-image-repeat",
    "border-image-slice",
    "border-image-source",
    "border-image-width",
    "border-inline",
    "border-inline-color",
    "border-inline-end",
    "border-inline-end-color",
    "border-inline-end-style",
    "border-inline-end-width",
    "border-inline-start",
    "border-inline-start-color",
    "border-inline-start-style",
    "border-inline-start-width",
    "border-inline-style",
    "border-inline-width",
    "border-left",
    "border-left-color",
    "border-left-style",
    "border-left-width",
    "border-radius",
    "border-right",
    "border-right-color",
    "border-right-style",
    "border-right-width",
    "border-spacing",
    "border-start-end-radius",
    "border-start-start-radius",
    "border-style",
    "border-top",
    "border-top-color",
    "border-top-left-radius",
    "border-top-right-radius",
    "border-top-style",
    "border-top-width",
    "border-width",
    "bottom",
    "box-decoration-break",
    "box-shadow",
    "box-sizing",
    "box-snap",
    "break-after",
    "break-before",
    "break-inside",
    "caption-side",
    "caret",
    "caret-color",
    "caret-shape",
    "chains",
    "clear",
    "clip",
    "clip-path",
    "clip-rule",
    "color",
    "color-adjust",
    "color-interpolation-filters",
    "color-scheme",
    "column-count",
    "column-fill",
    "column-gap",
    "column-rule",
    "column-rule-color",
    "column-rule-style",
    "column-rule-width",
    "columns",
    "column-span",
    "column-width",
    "contain",
    "content",
    "continue",
    "counter-increment",
    "counter-reset",
    "counter-set",
    "cue",
    "cue-after",
    "cue-before",
    "cursor",
    "direction",
    "display",
    "dominant-baseline",
    "elevation",
    "empty-cells",
    "fill",
    "fill-break",
    "fill-color",
    "fill-image",
    "fill-opacity",
    "fill-origin",
    "fill-position",
    "fill-repeat",
    "fill-rule",
    "fill-size",
    "filter",
    "flex",
    "flex-basis",
    "flex-direction",
    "flex-flow",
    "flex-grow",
    "flex-shrink",
    "flex-wrap",
    "float",
    "float-defer",
    "float-offset",
    "float-reference",
    "flood-color",
    "flood-opacity",
    "flow",
    "flow-from",
    "flow-into",
    "font",
    "font-family",
    "font-feature-settings",
    "font-kerning",
    "font-language-override",
    "font-optical-sizing",
    "font-palette",
    "font-size",
    "font-size-adjust",
    "font-stretch",
    "font-style",
    "font-synthesis",
    "font-synthesis-small-caps",
    "font-synthesis-style",
    "font-synthesis-weight",
    "font-variant",
    "font-variant-alternates",
    "font-variant-caps",
    "font-variant-east-asian",
    "font-variant-emoji",
    "font-variant-ligatures",
    "font-variant-numeric",
    "font-variant-position",
    "font-variation-settings",
    "font-weight",
    "footnote-display",
    "footnote-policy",
    "forced-color-adjust",
    "gap",
    "glyph-orientation-vertical",
    "grid",
    "grid-area",
    "grid-auto-columns",
    "grid-auto-flow",
    "grid-auto-rows",
    "grid-column",
    "grid-column-end",
    "grid-column-start",
    "grid-row",
    "grid-row-end",
    "grid-row-start",
    "grid-template",
    "grid-template-areas",
    "grid-template-columns",
    "grid-template-rows",
    "hanging-punctuation",
    "height",
    "hyphenate-character",
    "hyphenate-limit-chars",
    "hyphenate-limit-last",
    "hyphenate-limit-lines",
    "hyphenate-limit-zone",
    "hyphens",
    "image-orientation",
    "image-rendering",
    "image-resolution",
    "initial-letters",
    "initial-letters-align",
    "initial-letters-wrap",
    "inline-size",
    "inline-sizing",
    "inset",
    "inset-after",
    "inset-before",
    "inset-block",
    "inset-block-end",
    "inset-block-start",
    "inset-end",
    "inset-inline",
    "inset-inline-end",
    "inset-inline-start",
    "inset-start",
    "isolation",
    "justify-content",
    "justify-items",
    "justify-self",
    "leading-trim",
    "leading-trim-over",
    "leading-trim-under",
    "left",
    "letter-spacing",
    "lighting-color",
    "line-break",
    "line-clamp",
    "line-grid",
    "line-height",
    "line-height-step",
    "line-padding",
    "line-sizing",
    "line-snap",
    "list-style",
    "list-style-image",
    "list-style-position",
    "list-style-type",
    "margin",
    "margin-block",
    "margin-block-end",
    "margin-block-start",
    "margin-bottom",
    "margin-break",
    "margin-inline",
    "margin-inline-end",
    "margin-inline-start",
    "margin-left",
    "margin-right",
    "margin-top",
    "margin-trim",
    "marker",
    "marker-end",
    "marker-knockout-left",
    "marker-knockout-right",
    "marker-mid",
    "marker-pattern",
    "marker-segment",
    "marker-side",
    "marker-start",
    "mask",
    "mask-border",
    "mask-border-mode",
    "mask-border-outset",
    "mask-border-repeat",
    "mask-border-slice",
    "mask-border-source",
    "mask-border-width",
    "mask-clip",
    "mask-composite",
    "mask-image",
    "mask-mode",
    "mask-origin",
    "mask-position",
    "mask-repeat",
    "mask-size",
    "mask-type",
    "max-block-size",
    "max-height",
    "max-inline-size",
    "max-lines",
    "max-width",
    "min-block-size",
    "min-height",
    "min-inline-size",
    "min-width",
    "mix-blend-mode",
    "nav-down",
    "nav-left",
    "nav-right",
    "nav-up",
    "object-fit",
    "object-position",
    "offset",
    "offset-after",
    "offset-anchor",
    "offset-before",
    "offset-distance",
    "offset-end",
    "offset-path",
    "offset-position",
    "offset-rotate",
    "offset-start",
    "opacity",
    "order",
    "orphans",
    "outline",
    "outline-color",
    "outline-offset",
    "outline-style",
    "outline-width",
    "overflow",
    "overflow-anchor",
    "overflow-block",
    "overflow-inline",
    "overflow-wrap",
    "overflow-x",
    "overflow-y",
    "overscroll-behavior",
    "overscroll-behavior-block",
    "overscroll-behavior-inline",
    "overscroll-behavior-x",
    "overscroll-behavior-y",
    "padding",
    "padding-block",
    "padding-block-end",
    "padding-block-start",
    "padding-bottom",
    "padding-inline",
    "padding-inline-end",
    "padding-inline-start",
    "padding-left",
    "padding-right",
    "padding-top",
    "page",
    "page-break-after",
    "page-break-before",
    "page-break-inside",
    "pause",
    "pause-after",
    "pause-before",
    "perspective",
    "perspective-origin",
    "pitch",
    "pitch-range",
    "place-content",
    "place-items",
    "place-self",
    "play-during",
    "position",
    "quotes",
    "region-fragment",
    "resize",
    "rest",
    "rest-after",
    "rest-before",
    "richness",
    "right",
    "rotate",
    "row-gap",
    "ruby-align",
    "ruby-merge",
    "ruby-overhang",
    "ruby-position",
    "running",
    "scale",
    "scrollbar-color",
    "scrollbar-gutter",
    "scrollbar-width",
    "scroll-behavior",
    "scroll-margin",
    "scroll-margin-block",
    "scroll-margin-block-end",
    "scroll-margin-block-start",
    "scroll-margin-bottom",
    "scroll-margin-inline",
    "scroll-margin-inline-end",
    "scroll-margin-inline-start",
    "scroll-margin-left",
    "scroll-margin-right",
    "scroll-margin-top",
    "scroll-padding",
    "scroll-padding-block",
    "scroll-padding-block-end",
    "scroll-padding-block-start",
    "scroll-padding-bottom",
    "scroll-padding-inline",
    "scroll-padding-inline-end",
    "scroll-padding-inline-start",
    "scroll-padding-left",
    "scroll-padding-right",
    "scroll-padding-top",
    "scroll-snap-align",
    "scroll-snap-stop",
    "scroll-snap-type",
    "shape-image-threshold",
    "shape-inside",
    "shape-margin",
    "shape-outside",
    "spatial-navigation-action",
    "spatial-navigation-contain",
    "spatial-navigation-function",
    "speak",
    "speak-as",
    "speak-header",
    "speak-numeral",
    "speak-punctuation",
    "speech-rate",
    "stress",
    "string-set",
    "stroke",
    "stroke-align",
    "stroke-alignment",
    "stroke-break",
    "stroke-color",
    "stroke-dashadjust",
    "stroke-dasharray",
    "stroke-dash-corner",
    "stroke-dashcorner",
    "stroke-dash-justify",
    "stroke-dashoffset",
    "stroke-image",
    "stroke-linecap",
    "stroke-linejoin",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-origin",
    "stroke-position",
    "stroke-repeat",
    "stroke-size",
    "stroke-width",
    "table-layout",
    "tab-size",
    "text-align",
    "text-align-all",
    "text-align-last",
    "text-combine-upright",
    "text-decoration",
    "text-decoration-color",
    "text-decoration-line",
    "text-decoration-skip",
    "text-decoration-skip-box",
    "text-decoration-skip-ink",
    "text-decoration-skip-inset",
    "text-decoration-skip-self",
    "text-decoration-skip-spaces",
    "text-decoration-style",
    "text-decoration-thickness",
    "text-decoration-width",
    "text-emphasis",
    "text-emphasis-color",
    "text-emphasis-position",
    "text-emphasis-skip",
    "text-emphasis-style",
    "text-group-align",
    "text-indent",
    "text-justify",
    "text-orientation",
    "text-overflow",
    "text-shadow",
    "text-space-collapse",
    "text-space-trim",
    "text-spacing",
    "text-transform",
    "text-underline-offset",
    "text-underline-position",
    "text-wrap",
    "top",
    "transform",
    "transform-box",
    "transform-origin",
    "transform-style",
    "transition",
    "transition-delay",
    "transition-duration",
    "transition-property",
    "transition-timing-function",
    "translate",
    "unicode-bidi",
    "user-select",
    "vertical-align",
    "visibility",
    "voice-balance",
    "voice-duration",
    "voice-family",
    "voice-pitch",
    "voice-range",
    "voice-rate",
    "voice-stress",
    "voice-volume",
    "volume",
    "white-space",
    "widows",
    "width",
    "will-change",
    "word-boundary-detection",
    "word-boundary-expansion",
    "word-break",
    "word-spacing",
    "word-wrap",
    "wrap-after",
    "wrap-before",
    "wrap-flow",
    "wrap-inside",
    "wrap-through",
    "writing-mode",
    "z-index",
)



# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module provides the Theme class, which provides text formatting
properties based on the action (standard action) of a Token.

These properties can be used to colorize text according to a language
definition.

By default, the properties are read from a normal CSS (Cascading StyleSheets)
file and presented to the user of the Theme module through TextFormat objects,
although other storage backends could be devised.

A Theme provides a ``textformat()`` for standard actions, and for three
general situations: ``window()``, which denotes an editor window (or an
encompassing DIV or PRE block in HTML), ``selection()``, which is used for
selected text, and ``currentline()``, which can highlight the current line
the cursor is in in an editor.

From the TextFormat returned by ``selection()`` and ``currentline()``, in
most cases only the background color will be used.

The methods ``window()``, ``selection()`` and ``currentline()`` also accept a
state argument, which can be "default", "focus" or "disabled". A theme always
supports the default state, but can provide separate colors for the "focus"
or "disabled" state, which can be used to change the basic formatting in an
editor window based on its state (in keyboard focus or disabled). If a theme
does not support the "disabled" and/or "focus" state, the default scheme is
used.

In the ``themes/`` directory are bundled CSS themes that can be used.
Instantiate a bundled theme with::

    >>> from parce.theme import Theme
    >>> th = Theme.byname("default")

To use a custom CSS theme file, load it using::

    >>> th = Theme('/path/to/my/custom.css')

Get a TextFormat for an action, use e.g.::

    >>> f = th.textformat(String)
    >>> f
    <TextFormat color=Color(r=192, g=0, b=0, a=255)>


Mapping actions to CSS classes
------------------------------

Standard actions are mapped to a tuple of CSS class names: the action itself
and the actions it descends from. All CSS rules are combined, the one with the
most matches comes first.

For example, Comment maps to the "comment" CSS class, and Number maps
to ("literal", "number") because Number is a descendant action of Literal.

Some actions might have the same name, e.g. Escape and String.Escape.
Both match CSS rules with the ``.escape`` class selector, but a rule
with ``.string.escape`` will have higher precedence.

The order of the action names does not matter. E.g. an action Text.Comment
will match exactly the same CSS rules as an action Comment.Text. So you
should take some care when designing you action hierachy and not add too much
base action types.

"""


import collections
import itertools
import functools

from . import css
from . import themes
from . import util


class Theme:
    """A Theme maps a StandardAction to a TextFormat with CSS properties."""
    def __init__(self, filename="", stylesheet=""):
        """Instantiate Theme from a CSS file or text.

        If the ``stylesheet`` text is given, it is used as the stylesheet. Only
        if the text is empty, the filename is used to read the stylesheet from.

        """
        self._filename = filename
        self._css_text = stylesheet
        self.TextFormat = TextFormat

    @classmethod
    def byname(cls, name="default"):
        """Create Theme by name.

        The name is a CSS file in the themes/ directory, without the ".css"
        extension.

        """
        return cls(themes.filename(name))

    @util.cached_property
    def _stylesheet(self):
        """Load and cache the StyleSheet."""
        if self._css_text or not self._filename:
            return css.StyleSheet.from_text(self._css_text, self._filename)
        else:
            return css.StyleSheet.from_file(self._filename)

    @util.cached_property
    def style(self):
        """The stylesheet style rules (see :py:class:`css.Style <parce.css.Style>`)."""
        return self._stylesheet.style

    def filenames(self):
        """Return the list of filenames of the used stylesheet when instantiated"""
        return self._stylesheet.filenames()

    @util.cached_method
    def baseformat(self, role="window", state="default"):
        """Return a TextFormat for a specific role and a state.

        The role may be any string that maps to a CSS class in the theme
        CSS file that is available there together with the 'parce' class.

        The following roles are recognized and used by parce, but you may
        also define your own roles in your (applications') theme CSS files:

        ``"window"``
            The TextFormat for the editor window or the encompassing DIV when
            formatting HTML. Corresponds to the "parce" CSS class alone in the
            theme file. You can set color, background and, if desired, font
            preferences.
        ``"selection"``
            The TextFormat to use for selected text. Uses the ``::selection``
            pseudo element.
        ``"current-line"``
            The TextFormat for the current line. If you use it, set only the
            *background* color in your theme file.
        ``"trialing-whitespace"``
            The TextFormat (use only the *background*) to highlight trialing
            whitespace, if desired.
        ``"eol-marker"``
            The *color* to draw a "end-of-line" marker with, if desired

        The state argument may be "default", "focus", or "disabled", and
        reflects the state of the user interface the style variant is used for.

        """
        if role == "window":
            e = css.Element(class_="parce")
        elif role == "selection":
            e = css.Element(class_="parce", pseudo_elements=["selection"])
        else:
            e = css.Element(class_="parce {}".format(role))
        if state in ("focus", "disabled"):
            e.pseudo_classes = [state]
        return self.TextFormat(self.style.select_element(e).properties())

    @util.cached_method
    def textformat(self, action):
        """Return the TextFormat for the specified action."""
        class_ = repr(action).lower().replace('.', ' ')
        e = css.Element(class_=class_, parent=css.Element(class_="parce"))
        return self.TextFormat(self.style.select_element(e).properties())

    def tokens(self, theme_context, tree, start=0, end=None):
        """Yield tokens from ``tree.tokens_range(start, end)``.

        If a context is entered, ``theme_context.push(language)`` is called
        with the language of the context's lexicon. If the context is left,
        ``theme_context.pop()`` is called.

        In Theme, the ``theme_context`` argument is unused.

        """
        yield from tree.tokens_range(start, end)


class MetaTheme:
    """A Theme that encapsulates a default Theme and per-language sub-Themes."""
    def __init__(self, theme):
        """Instantiate with default theme."""
        self._theme = theme
        self._themes = {}
        self._add_window = {theme: False}

    def add_theme(self, language, theme, add_window=False):
        """Add a specific Theme for the specified Language.

        If ``add_window`` is set to True, the formatter will render text from
        the specified language with its own ``window`` default style added to
        all formats. A formatter will also need to take care to render
        untokenized ranges with the ``window()`` style of the sub-theme.

        """
        self._themes[language] = theme
        self._add_window[theme] = add_window

    def get_theme(self, language):
        """Return the tuple(theme, add_window) for the language.

        If the exact language was not found, the base classes of the Language are
        tried. If still no luck, self is returned.

        """
        try:
            return self._themes[language]
        except KeyError:
            pass
        for lang in language.mro()[:-2]:    # don't test Language and object
            try:
                return self._themes[lang]
            except KeyError:
                pass
        return self

    def get_add_window(self, theme):
        """Return the ``add_window`` value that was set when adding this sub-theme."""
        return self._add_window.get(theme, False)

    def baseformat(self, role="window", state="default"):
        """Return the base TextFormat for the rold and state of the default theme."""
        return self._theme.baseformat(state)

    def textformat(self, action):
        """Return the TextFormat for the specified action of the default theme."""
        return self._theme.textformat(action)

    def tokens(self, theme_context, tree, start=0, end=None):
        """Yield tokens from ``tree.tokens_range(start, end)``.

        If a context is entered, ``theme_context.push(language)`` is called
        with the language of the context's lexicon. If the context is left,
        ``theme_context.pop()`` is called.

        The theme_context can switch the theme, based on the current language.

        """
        for context, slice_ in tree.context_slices(start, end):
            theme_context.push(context.lexicon.language)
            for n in context[slice_]:
                stack = []
                i = 0
                while True:
                    for i in range(i, len(n)):
                        m = n[i]
                        if m.is_token:
                            yield m
                        else:
                            stack.append(i)
                            i = 0
                            n = m
                            theme_context.push(n.lexicon.language)
                            break
                    else:
                        if stack:
                            theme_context.pop()
                            n = n.parent
                            i = stack.pop() + 1
                        else:
                            break
            theme_context.pop()


class TextFormat:
    """Simple textformat that reads CSS properties and supports a subset of those.

    This factory is used by default by Theme, but you can implement your own.
    Such a factory only needs to implement an ``__init__`` method that reads
    the dictionary of property Value lists returned by Style.properties().

    A TextFormat has a False boolean value if no single property is set.

    You can add and subtract TextFormats::

        >>> import parce
        >>> t = parce.theme_by_name()
        >>> f = t.window()
        >>> f
        <TextFormat background_color=Color(r=255, g=255, b=240, a=1.0), color=
        Color(r=0, g=0, b=0, a=1.0), font_family=['monospace'], font_size=12,
        font_size_unit='pt'>
        >>> f2 = t.textformat(parce.Comment)
        >>> f2
        <TextFormat color=Color(r=105, g=105, b=105, a=1.0), font_family=['serif'],
        font_style='italic'>
        >>> f + f2
        <TextFormat background_color=Color(r=255, g=255, b=240, a=1.0), color=
        Color(r=105, g=105, b=105, a=1.0), font_family=['serif'], font_size=12,
        font_size_unit='pt', font_style='italic'>
        >>> f - f2
        <TextFormat background_color=Color(r=255, g=255, b=240, a=1.0), color=
        Color(r=0, g=0, b=0, a=1.0), font_family=['monospace'], font_size=12,
        font_size_unit='pt'>
        >>>

    Adding a TextFormat returns a new format with our properties set and then
    the properties of the other. This is useful when it is not possible to
    overlay properties with underlying window properties.

    Subtracting a TextFormat returns a new format with the properties removed
    that are the same in the other format. This is useful when properties of
    a certain action happen to be the same as the underlying window properties;
    it is not needed to set these again in such cases.

    """
    color = None                    #: the foreground color as Color(r, g, b, a) tuple
    background_color = None         #: the background color (id)
    text_decoration_color = None    #: the color for text decoration
    text_decoration_line = ()       #: underline, overline and/or line-through
    text_decoration_style = None    #: solid, double, dotted, dashed or wavy
    font_family = ()                #: family or generic name
    font_kerning = None             #: font kerning
    font_size = None                #: font size
    font_size_unit = None           #: font size unit if given
    font_stretch = None             #: font stretch value (keyword or float, 1.0 is normal)
    font_style = None               #: normal, italic or oblique
    font_style_angle = None         #: oblique slant if given
    font_style_angle_unit = None    #: oblique slant unit if given
    font_variant_caps = None        #: all kind of small caps
    font_variant_position = None    #: normal, sub or super
    font_weight = None              #: 100 - 900 or keyword like ``bold``

    dispatch = util.Dispatcher()

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__,
            ", ".join("{}={}".format(key, repr(value))
                                for key, value in sorted(self.__dict__.items())))

    def __init__(self, properties):
        for prop, values in properties.items():
            self.dispatch(prop, values)

    def __bool__(self):
        """Return True if at least one property is set."""
        return bool(self.__dict__)

    def __sub__(self, other):
        """Return a new TextFormat with the properties removed that are the same in ``other``."""
        new = type(self)({})
        for props in (
            ('color',),
            ('background_color',),
            ('text_decoration_color', 'text_decoration_line', 'text_decoration_style'),
            ('font_family',),
            ('font_kerning',),
            ('font_size', 'font_size_unit'),
            ('font_stretch',),
            ('font_style', 'font_style_angle', 'font_style_angle_unit'),
            ('font_variant_caps',),
            ('font_variant_position',),
            ('font_weight',),
        ):
            if any(prop in self.__dict__ for prop in props) and \
                    any(self.__dict__.get(prop) != other.__dict__.get(prop) for prop in props):
                for prop in props:
                    try:
                        new.__dict__[prop] = self.__dict__[prop]
                    except KeyError:
                        pass
        return new

    def __add__(self, other):
        """Return a new TextFormat adding the other's properties."""
        new = type(self)({})
        new.__dict__.update(self.__dict__)
        new.__dict__.update(other.__dict__)
        return new

    def __eq__(self, other):
        """Return True if other has the same properties."""
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Return True if other has different properties."""
        return self.__dict__ != other.__dict__

    def css_properties(self):
        """Return a dict usable to write out a CSS rule with our properties."""
        return dict(itertools.chain(
            self.write_color(),
            self.write_text_decoration(),
            self.write_font(),
        ))

    def write_color(self):
        """Yield color and background color as CSS properties, if set."""
        if self.color:
            yield "color", css.color2hex(self.color)
        if self.background_color:
            yield "background-color", css.color2hex(self.background_color)

    def write_text_decoration(self):
        """Yield a text-decoration property, if set."""
        props = []
        props.extend(self.text_decoration_line)
        if self.text_decoration_style:
            props.append(self.text_decoration_style)
        if self.text_decoration_color:
            props.append(css.color2hex(self.text_decoration_color))
        if props:
            yield "text-decoration", " ".join(props)

    def write_font(self):
        """Yield all font-xxxx properties, if set."""
        if self.font_family:
            yield "font-family", ", ".join(map(css.quote_if_needed, self.font_family))
        if self.font_size:
            yield "font-size", "{}{}".format(
                self.font_size, self.font_size_unit or "")
        if self.font_style == "oblique" and self.font_style_angle:
            slant = self.font_style_angle
            unit = self.font_style_angle_unit or ""
            yield "font-style", "oblique {}{}".format(slant, unit)
        elif self.font_style:
            yield "font-style", self.font_style
        if self.font_stretch:
            yield "font-stretch", format(self.font_stretch)
        if self.font_kerning:
            yield "font-kerning", self.font_kerning
        if self.font_variant_caps:
            yield "font-variant-caps", self.font_variant_caps
        if self.font_variant_position:
            yield "font-variant-position", self.font_variant_position
        if self.font_weight:
            yield "font-weight", format(self.font_weight)

    @dispatch("color")
    def read_color(self, values):
        for v in values:
            if v.color:
                self.color = v.color
                return

    @dispatch("background-color")
    def read_background_color(self, values):
        for v in values:
            if v.color:
                self.background_color = v.color
                return

    @dispatch("background")
    def read_background(self, values):
        self.read_background_color(values)

    @dispatch("text-decoration-color")
    def read_text_decoration_color(self, values):
        for v in values:
            if v.color:
                self.text_decoration_color = v.color
                return

    @dispatch("text-decoration-line")
    def read_text_decoration_line(self, values):
        decos = []
        for v in values:
            if v.text in ("underline", "overline", "line-through"):
                decos.append(v.text)
            elif v.text == "none":
                decos.clear()
        self.text_decoration_line = decos

    @dispatch("text-decoration-style")
    def read_text_decoration_style(self, values):
        for v in values:
            if v.text in ("solid", "double", "dotted", "dashed", "wavy"):
                self.text_decoration_style = v.text
                return

    @dispatch("text-decoration")
    def read_text_decoration(self, values):
        self.read_text_decoration_color(values)
        self.read_text_decoration_line(values)
        self.read_text_decoration_style(values)

    @dispatch("font-family")
    def read_font_family(self, values):
        families = []
        for v in values:
            if v.text and (v.quoted or v.text in (
                "serif",
                "sans-serif",
                "monospace",
                "cursive",
                "fantasy",
                "system-ui",
                "math",
                "emoji",
                "fangsong",
            )):
                families.append(v.text)
        self.font_family = families

    @dispatch("font-kerning")
    def read_font_kerning(self, values):
        for v in values:
            if v.text in ("auto", "normal", "none"):
                self.font_kerning = v.text
                return

    @dispatch("font-size")
    def read_font_size(self, values):
        for v in values:
            if v.text in ("xx-small", "x-small", "small", "medium",
                          "large", "x-large", "xx-large", "xxx-large",
                          "larger", "smaller"):
                self.font_size = v.text
                return
            elif v.number is not None:
                self.font_size = v.number
                self.font_size_unit = v.unit
                return

    @dispatch("font-stretch")
    def read_font_stretch(self, values):
        for v in values:
            if v.text in ("ultra-condensed", "extra-condensed", "condensed",
                          "semi-condensed", "semi-expanded", "expanded",
                          "extra-expanded", "ultra-expanded"):
                self.font_stretch = v.text
            elif v.number is not None and v.unit == "%":
                self.font_stretch = v.number

    @dispatch("font-style")
    def read_font_style(self, values):
        v = values[0]
        for n in values[1:] + [None]:
            if v.text in ("normal", "italic"):
                self.font_style = v.text
                return
            elif v.text == "oblique":
                self.font_style = v.text
                if n and n.number is not None:
                    self.font_style_angle = n.number
                    self.font_style_angle_unit = n.unit
                    return
            v = n

    @dispatch("font-variant-caps")
    def read_font_variant_caps(self, values):
        for v in values:
            if v.text in ("normal", "small-caps", "all-small-caps", "petite-caps",
                          "all-petite-caps", "unicase", "titling-caps"):
                self.font_variant_caps = v.text
                return

    @dispatch("font-variant-position")
    def read_font_variant_position(self, values):
        for v in values:
            if v.text in ("normal", "sub", "super"):
                self.font_variant_position = v.text
                return

    @dispatch("font-weight")
    def read_font_weight(self, values):
        for v in values:
            if v.text in ("normal", "bold", "lighter", "bolder"):
                self.font_weight = v.text
                return
            elif v.number is not None:
                self.font_weight = v.number
                return

    @dispatch("font")
    def read_font(self, values):
        self.read_font_style(values)
        numvalues = []
        for v in values:
            if v.text in ("caption", "icon", "menu", "message-box",
                          "small-caption", "status-bar"):
                self.font_family = [v.text]
                return
            elif v.text in ("normal", "small-caps"):
                self.font_variant_caps = v.text
            elif v.text in ("ultra-condensed", "extra-condensed", "condensed",
                            "semi-condensed", "semi-expanded", "expanded",
                            "extra-expanded", "ultra-expanded"):
                self.font_stretch = v.text
            elif v.text in ("bold", "lighter", "bolder"):
                self.font_weight = v.text
            elif v.number is not None:
                numvalues.append((v.number, v.unit))
        self.read_font_family(values)
        # if more than one size was given, weight is the first
        if len(numvalues) == 1:
            self.font_size, self.font_size_unit = numvalues[0]
        elif len(numvalues) > 1:
            if self.font_weight is None:
                self.font_weight = numvalues[0][0]
            self.font_size, self.font_size_unit = numvalues[1]





# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Test various transformations.
"""

import sys
sys.path.insert(0, '.')

import parce.transform


def transform_file(filename, root_lexicon=None, transform=None):
    s = open(filename).read()
    if root_lexicon is None:
        root_lexicon = parce.find(filename=filename, contents=s)
    return parce.transform.transform_text(root_lexicon, s, transform)

JSON_RESULT = \
{'background': 'background.png',
 'comment': 'JSON example',
 'contents': [{'path': '/Applications', 'type': 'link', 'x': 449, 'y': 320},
              {'path': '../dist/Frescobaldi.app',
               'type': 'file',
               'x': 188,
               'y': 320},
              {'path': '../README.txt', 'type': 'file', 'x': 100, 'y': 70},
              {'path': '../ChangeLog.txt', 'type': 'file', 'x': 100, 'y': 185},
              {'path': '../COPYING.txt', 'type': 'file', 'x': 540, 'y': 70}],
 'icon-size': 80,
 'title': 'Frescobaldi'}



def test_main():

    result = transform_file('tests/lang/example.json')
    assert result == JSON_RESULT


if __name__ == "__main__":
    test_main()




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Parse JavaScript.

"""

__all__ = ('JavaScript',)

import re

from parce import Language, lexicon, skip, default_action
from parce.rule import TEXT, MATCH, arg, bygroup, call, dselect, select, words
from parce.action import (
    Bracket, Comment, Delimiter, Keyword, Literal, Name, Number, Operator,
    Separator, String)

from parce.unicharclass import categories
from . import javascript_words as js


RE_JS_IDENT_STARTCHAR = r'$_' + ''.join(map(categories.get, ['Lu', 'Ll', 'Lt', 'Lm', 'Lo', 'Nl']))
RE_JS_IDENT_CHAR = RE_JS_IDENT_STARTCHAR + '\u200c\u200d' + ''.join(map(categories.get, ['Mn', 'Mc', 'Nd', 'Pc']))
RE_JS_ESCAPE_CHAR = r'\\u[0-9a-fA-F]{4}'
RE_JS_IDENT_TOKEN = _I_ = fr'(?:[{RE_JS_IDENT_STARTCHAR}]|{RE_JS_ESCAPE_CHAR})' \
                fr'(?:[{RE_JS_IDENT_CHAR}]+|{RE_JS_ESCAPE_CHAR})*'

RE_JS_DECIMAL_NUMBER = r'\d+(?:_\d+)*n|(?:\.\d+(?:_\d+)*|\d+(?:_\d+)*(?:\.(?:\d+(?:_\d+)*)?)?)(?:[eE][-+]\d+(?:_\d+)*)?'
RE_JS_REGEXP = r'/[^*/\n](?:\\[\[\\^$.|?*+()]|\[(?:\\[\\\[\]]|[^\]])+\]|[^/\[\n])*/[gimsuy]?'


class JavaScript(Language):
    @lexicon
    def root(cls):
        yield r"'", String.Start, cls.string("'")
        yield r'"', String.Start, cls.string('"')
        yield r'`', String.Start, cls.template_literal
        yield '//', Comment, cls.singleline_comment
        yield r'/\*', Comment.Start, cls.multiline_comment
        yield fr'(const|let|var)\s+({_I_})\b', bygroup(Keyword, Name.Variable.Definition)
        yield fr'(function)\s+({_I_})\b', bygroup(Keyword, Name.Function.Definition)
        yield fr'(new)\s+({_I_})\b', bygroup(Keyword, Name.Class.Definition)
        yield words(js.JAVASCRIPT_KEYWORDS, prefix=r'\b', suffix=r'\b'), Keyword
        yield words(js.JAVASCRIPT_DECLARATORS, prefix=r'\b', suffix=r'\b'), Keyword
        yield words(js.JAVASCRIPT_RESERVED_KEYWORDS, prefix=r'\b', suffix=r'\b'), Keyword.Reserved
        yield words(js.JAVASCRIPT_CONSTANTS, prefix=r'\b', suffix=r'\b'), Name.Constant
        yield words(js.JAVASCRIPT_BUILTINS, prefix=r'\b', suffix=r'\b'), Name.Builtin
        yield words(js.JAVASCRIPT_PROTOTYPES, prefix=r'\b', suffix=r'\b'), Name.Builtin
        yield fr'(\.)\s*({_I_})\b(?:\s*([\(\[]))?', bygroup(Delimiter,
                dselect(MATCH(3), {'(': Name.Method}, Name.Attribute), Delimiter), \
            dselect(MATCH(3), {'(': cls.call, '[': cls.index})
        yield fr'({_I_})\b(?:\s*([\(\[]))?', bygroup(
                dselect(MATCH(2), {'(': Name.Function}, Name.Variable), Delimiter), \
            dselect(MATCH(2), {'(': cls.call, '[': cls.index})
        yield fr'{_I_}\b', select(call(str.isupper, TEXT), Name.Variable, Name.Class)
        ## numerical values (recently, underscore support inside numbers was added)
        yield '0[oO](?:_?[0-7])+n?', Number
        yield '0[bB](?:_?[01])+n?', Number
        yield '0[xX](?:_?[0-9a-fA-F])+n?', Number
        yield RE_JS_DECIMAL_NUMBER, Number
        yield r'\{', Bracket.Start, cls.scope
        yield r'\[', Bracket.Start, cls.array
        yield r'\(', Delimiter, cls.paren
        yield RE_JS_REGEXP, Literal.Regexp
        yield r'(?:<<|>>>?|[&|^*/%+-])=', Operator.Assignment
        yield r'&&?|\|\|?|<<|>>>?|[!=]==?|<=?|>=?|\*\*|[-+~!/*%^?:,]', Operator
        yield r'=', Operator.Assignment
        yield r';', Delimiter

    @lexicon
    def scope(cls):
        yield r'\}', Bracket.End, -1
        yield from cls.root

    @lexicon
    def call(cls):
        """name(...) syntax."""
        yield r'\)', Delimiter, -1
        yield from cls.root

    @classmethod
    def expression(cls):
        """Stuff between ( ) or [ ]"""
        yield r'\{', Bracket.Start, cls.object
        yield from cls.root

    @lexicon
    def object(cls):
        """An object (dictionary) { ... }."""
        yield r'[:,]', Separator
        yield r'\}', Bracket.End, -1
        yield from cls.expression()

    @lexicon
    def array(cls):
        """An array [ ... ]."""
        yield r',', Separator
        yield r'\]', Bracket.End, -1
        yield from cls.expression()

    @lexicon
    def paren(cls):
        """An expression between ( ... )."""
        yield r',', Separator
        yield r'\)', Delimiter, -1
        yield from cls.expression()

    @lexicon
    def index(cls):
        """name[...] syntax."""
        yield r'\]', Delimiter, -1
        yield from cls.root

    @lexicon
    def string(cls):
        yield arg(), String.End, -1
        yield (r'''\\(?:[0"'\\nrvtbf]'''
            r'|x[a-fA-F0-9]{2}'
            r'|u[a-fA-F0-9]{4}'
            r'|u\{[a-fA-F0-9]{1,6}\})'), String.Escape
        yield default_action, String

    @lexicon
    def template_literal(cls):
        yield from cls.string('`')
        yield r'\\[$`]', String.Escape
        yield r'\$\{', Delimiter.Template, cls.template_literal_expression

    @lexicon
    def template_literal_expression(cls):
        yield r'\}', Delimiter.Template, -1
        yield from cls.root


    #------------------ comments -------------------------
    @lexicon(re_flags=re.MULTILINE)
    def singleline_comment(cls):
        yield '$', None, -1
        yield from cls.comment_common()

    @lexicon
    def multiline_comment(cls):
        yield r'\*/', Comment.End, -1
        yield from cls.comment_common()




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import collections
import re
import reprlib

import parce
from .lexicon import LexiconDescriptor, Lexicon
from .ruleitem import variations, a_number


def validate_language(lang):
    """Validate all lexicons in this language.

    Errors and warnings are printed to stdout. If there are errors,
    this function returns False, otherwise True.

    """

    lexicons = []
    for key, value in lang.__dict__.items():
        if isinstance(value, LexiconDescriptor):
            lexicons.append(getattr(lang, key))

    errors = set()
    warnings = set()
    for lexicon in lexicons:
        v = LexiconValidator(lexicon)
        v.validate()
        errors.update(v.errors)
        warnings.update(v.warnings)
    for warning in warnings:
        print(warning)
    for error in errors:
        print(error)
    return not errors


class LexiconValidator:

    def __init__(self, lexicon):
        self.lexicon = lexicon
        self.errors = set()
        self.warnings = set()

    def error(self, msg, lexicon=None):
        """Add message to errors."""
        msg = "{}: error: {}".format(lexicon or self.lexicon, msg)
        self.errors.add(msg)

    def warning(self, msg, lexicon=None):
        """Add message to warnings."""
        msg = "{}: warning: {}".format(lexicon or self.lexicon, msg)
        self.warnings.add(msg)

    def validate(self):
        """Validate a lexicon.

        Errors and warnings are left in the ``errors`` and ``warnings``
        attributes, respectively.

        """
        self.errors.clear()
        self.warnings.clear()
        default_act, default_tg = None, None
        patterns = set()
        for n, rule in enumerate(self.lexicon.rules, 1):
            if not isinstance(rule, (tuple, list)):
                self.error("rule #{}: invalid rule; should be tuple or list".format(n))
                continue
            elif len(rule) < 2:
                self.error("rule #{}: invalid rule; pattern and action should be there".format(n))
                continue
            pattern, *rule = rule
            if pattern is parce.default_action:
                if default_act:
                    self.error("rule #{}: conflicting default actions".format(n))
                else:
                    default_act = rule[0]
            elif pattern is parce.default_target:
                if default_tg:
                    self.error("rule #{}: conflicting default targets".format(n))
                else:
                    default_tg = rule
                    self.check_default_target(default_tg)
            else:
                if pattern is None:
                    self.warning("rule #{}: pattern is None; rule will be skipped".format(n))
                else:
                    self.validate_pattern(pattern, n)
                    if pattern in patterns:
                        self.warning("rule #{0}: repeated pattern {1}; will be skipped".format(n, reprlib.repr(pattern)))
                    patterns.add(pattern)
                self.validate_rule(rule, n)

        if default_act and default_tg:
            self.error("can't have both default_action and default_target")
        return not self.errors

    def validate_pattern(self, pattern, n):
        """Validate a regular expression pattern."""
        try:
            rx = re.compile(pattern, self.lexicon.re_flags)
        except (TypeError, re.error) as e:
            self.error("rule #{0}: regular expression {1} error:\n  {2}".format(n, repr(pattern), e))
        else:
            if rx.match(''):
                self.warning("rule #{0}: pattern {1} matches the empty string".format(n, repr(pattern)))

    def validate_rule(self, rule, n):
        """Validate a rule, which should be action, target[, target, ...].

        Does not look at the action, but checks whether all the targets are
        valid (either an integer or a Lexicon).

        """
        # all possible rule paths
        for path in variations(rule):
            for target in path[1:]:     # the first item always is the action
                if target is not a_number and not isinstance(target, (int, Lexicon)):
                    self.error("rule #{0}: invalid target: {1}".format(n, target))

    def check_default_target(self, target):
        """Check whether this default target could lead to circular references.

        This could hang the parser, and we wouldn't like to have that :-)

        """
        # a unique object for every entered context, mimicking tree builder behaviour
        class Context:
            def __init__(self, lexicon):
                self.lexicon = lexicon

        lexicon = self.lexicon
        lexicons = collections.Counter()    # count them to find the circular culprits
        state = [Context(lexicon)]
        circular = set()                    # track circular (existing) contexts
        warn = False
        context = state[-1]
        while True:
            circular.add(context)
            lexicons.update((lexicon,))     # count
            depth = len(state)
            for t in target:
                if isinstance(t, int):
                    if t < 0:
                        if len(state) + t < 1:
                            return
                        del state[t:]
                    else:
                        for _ in range(t):
                            state.append(Context(lexicon))
                elif not isinstance(t, Lexicon):
                    self.error("in default target only integer or lexicon allowed", lexicon)
                    return
                else:
                    state.append(Context(t))
            newcontext = state[-1]
            lexicon = newcontext.lexicon
            if newcontext is context:
                self.error("invalid default target", lexicon)
                return
            context = newcontext
            if newcontext in circular:
                # a circular default state that lands in an existing context
                # is handled gracefully by the tree builder
                warn = True
            if len(circular) > 100:
                lexicons = " <-> ".join(str(l) for l, n in lexicons.items() if n > 1)
                if warn:
                    # this type of circular default state is handled
                    self.warning("handled circular default target: {}".format(lexicons), lexicon)
                else:
                    # run away default states creating new contexts all the time
                    self.error("circular default target: {}".format(lexicons), lexicon)
                return
            for pattern, *target in lexicon():
                if pattern is parce.default_target:
                    break
            else:
                break




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
This module contains the implementation of the replacable rule item objects,
and some functions to query and manipulate rules that are used by the Lexicon.

"""


from . import util


# pre_evaluate flags
_CHANGED = 0
_COMPLETE = 1
_UNCHANGED = 2


class _EvaluationError(RuntimeError):
    """Raised when an attribute is missing in the evaluation namespace."""
    pass


class Item:
    """Base class for any replacable rule item.

    An Item is considered to be immutable; you should never alter the
    attributes after instantiation. When an Item can be partly pre-evaluated a
    copy must be returned by calling ``type(self)(*args)``.

    In some cases you can also return a different Item type when pre-evaluating
    partly succeeds. For example, the :class:`select` type simply returns the
    chosen item if the index already can be evaluated, without evaluating the
    other items.

    """
    __slots__ = ()

    def __getitem__(self, n):
        """Return a new Item that performs item[n]. n is evaluated as well."""
        return call(_get_item, self, n)

    def evaluate(self, ns):
        """Evaluate item in namespace dict ``ns``."""
        raise NotImplementedError

    def pre_evaluate(self, ns):
        """Try to evaluate item in namespace dict ``ns``.

        Return a two-tuple(obj, success).

        Success is a two-bit value indicating whether the result is completely
        evaluated and whether something has changed. Bit 0 is set when the item
        is completely evaluated, and bit 1 is set when there was no
        modification. So all possible return values are:

        | 0: the object has changed but it is not yet completely evaluated
        | 1: the object has changed and now it is fully evaluated
        | 2: the object needs evaluation but is not changed

        Specific items may reimplement this method to return partially
        evaluated items.

        """
        try:
            return self.evaluate(ns), _COMPLETE
        except _EvaluationError:
            return self, _UNCHANGED

    def variations(self):
        """Yield the possible results for this item.

        This is used to build a decision tree for a rule, to see which actions
        and targets it could bring.

        The default implementation raises a RuntimeError; only RuleItem
        objects can yield variations.

        """
        raise RuntimeError("Item '{}' can't be used directly in a rule".format(repr(self)))

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ', '.join(map(repr, self._repr_args())))

    def _repr_args(self):
        return ()


# helper function used for Item.__getitem__
def _get_item(text, n):
    return text[n]


class RuleItem(Item):
    """Base class for items that may become visible in rules."""
    __slots__ = ()


class VariableItem(Item):
    """A named variable that's accessed in the namespace.

    If a predicate is given, it is called when this item is called, with this
    item as first argument, and then the other arguments that were specified.

    """
    __slots__ = ('_name', '_predicate')
    def __init__(self, name, predicate=None):
        self._name = name
        self._predicate = predicate

    def __call__(self, *args):
        return call(self._predicate, self, *args)

    def evaluate(self, ns):
        """Get the variable from the namespace dict."""
        try:
            return ns[self._name]
        except KeyError as e:
            raise _EvaluationError("Can't find variable '{}'".format(self._name)) from e

    def __repr__(self):
        return self._name.upper()


class call(Item):
    """Call predicate with arguments."""
    __slots__ = ('_predicate', '_arguments')
    def __init__(self, predicate, *arguments):
        self._predicate = predicate
        self._arguments = arguments

    def evaluate(self, ns):
        """Call predicate with the arguments."""
        predicate = evaluate(self._predicate, ns)
        arguments = evaluate(self._arguments, ns)
        return predicate(*arguments)

    def pre_evaluate(self, ns):
        """Optimize by pre-evaluating what can be pre-evaluated."""
        predicate, pred_ok = pre_evaluate(self._predicate, ns)
        arguments, arg_ok  = pre_evaluate(self._arguments, ns)
        ok = pred_ok & arg_ok
        if ok & _COMPLETE:
            return predicate(*arguments), _COMPLETE
        if ok & _UNCHANGED:
            return self, _UNCHANGED
        return type(self)(predicate, *arguments), _CHANGED

    def _repr_args(self):
        return (self._predicate, *self._arguments)


class RuleItem(Item):
    """Classes inheriting RuleItem are allowed in toplevel in rules.

    They are evaluated by the lexicon when a rule matches.

    """
    __slots__ = ()


class select(RuleItem):
    """Chooses one of the items.

    If an item is a list, it is unrolled when replacing the item in a rule.

    """
    __slots__ = ('_index', '_items')

    def __init__(self, index, *items):
        self._index = index
        self._items = items

    def evaluate(self, ns):
        """Return items[index]."""
        index = evaluate(self._index, ns)
        item = evaluate(self._items[index], ns)
        return item

    def pre_evaluate(self, ns):
        """Optimize by pre-evaluating what can be pre-evaluated."""
        index, ok = pre_evaluate(self._index, ns)
        if ok & _COMPLETE:
            item, ok = pre_evaluate(self._items[index], ns)
            return item, ok & _COMPLETE     # mask unchanged state
        items, items_ok = pre_evaluate(self._items, ns)
        ok &= items_ok
        if ok & _UNCHANGED:
            return self, _UNCHANGED
        return type(self)(index, *items), _CHANGED

    def variations(self):
        """Yield all the items that could be chosen (unevaluated)."""
        yield from self._items

    def _repr_args(self):
        return (self._index, *self._items)


class target(RuleItem):
    """target(value, *lexicons)

    Has a special handling: if the value is an integer, it is used as the
    result value (to push/pop contexts).

    If it is a two-tuple(index, argument): The index points to the lexicon and
    the argument is used as lexicon argument.

    """
    __slots__ = ('_value', '_lexicons')

    def __init__(self, value, *lexicons):
        self._value = value
        self._lexicons = lexicons

    def evaluate(self, ns):
        """Return value if integer, otherwise lexicons[value[0]](value[1])."""
        value = evaluate(self._value, ns)
        if isinstance(value, int):
            return value
        index, arg = value
        lexicon = evaluate(self._lexicons[index], ns)
        return lexicon if arg is None else lexicon(arg)

    def pre_evaluate(self, ns):
        """Optimize by pre-evaluating what can be pre-evaluated."""
        value, ok = pre_evaluate(self._value, ns)
        if ok & _COMPLETE:
            if isinstance(value, int):
                return value, _COMPLETE
            index, arg = value
            lexicon, ok = pre_evaluate(self._lexicons[index], ns)
            if ok & _COMPLETE:
                return (lexicon if arg is None else lexicon(arg), _COMPLETE)
            return type(self)((0, arg), lexicon), _CHANGED
        # pre-evaluate the lexicons
        lexicons, l_ok = pre_evaluate(self._lexicons, ns)
        ok &= l_ok
        if ok & _UNCHANGED:
            return self, ok
        return type(self)(value, *lexicons), ok

    def variations(self):
        """Yield our possible variations.

        If the value is evaluated, yield either the value or the chosen
        lexicon. If not, yields ``a_number`` and all lexicons items.

        """
        value = self._value
        if isinstance(value, Item):
            yield a_number
            yield from self._lexicons
        elif isinstance(value, int):
            yield value
        else:
            index, arg = value
            yield self._lexicons[index]

    def _repr_args(self):
        return (self._value, *self._lexicons)


class PostponedItem(RuleItem):
    """Mixin base class for items that keep alive after the Lexicon.

    When inheriting from this class, implement the :meth:`evaluate_items`
    method, which lists all values as they were given to the __init__ method.

    If this method returns values, those are evaluated, and a new PostponedItem
    is returned with the contents evaluated.

    """
    __slots__ = ()

    def evaluate(self, ns):
        """Evaluate all values returned by the evaluate_items() method.

        If any value changes, a copy of the Item is returned, otherwise the
        Item ifself. If the evaluate_items() method does not yield any value,
        this Item is always returned unchanged.

        """
        items, ok = pre_evaluate(self.evaluate_items(), ns)
        return self if ok & _UNCHANGED else type(self)(*items)

    def pre_evaluate(self, ns):
        """Pre-evaluate all values returned by the evaluate_items() method.

        If any value changes, a copy of the Item is returned, otherwise the
        Item ifself.

        """
        items, ok = pre_evaluate(self.evaluate_items(), ns)
        return self if ok & _UNCHANGED else type(self)(*items), ok

    def evaluate_items(self):
        """Return a tuple of the values as given to the __init__ method,
        when they need to be evaluated inside this PostponedItem.

        This method should either yield *all* values that were given to the
        __init__ method, or nothing. The default implementation yields nothing,
        so nothing is evaluated or pre-evaluated.

        """
        return ()


class pattern(PostponedItem):
    """Represents a pattern.

    This evaluates its value, but remains alive after building the rule.

    """
    __slots__ = ('_value',)

    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        """Get the pattern value."""
        return self._value

    def evaluate_items(self):
        """Yield the pattern value."""
        return self._value,

    def variations(self):
        """If the value is evaluated, yield it, otherwise yields ``None`` and ``a_string``."""
        if isinstance(self._value, Item):
            yield None
            yield a_string
        else:
            yield self._value

    def _repr_args(self):
        return self._value,


class ActionItem(PostponedItem):
    """Mixin base class for dynamic actions."""
    __slots__ = ()


class SubgroupAction(ActionItem):
    """Yield actions from subgroups in a match.

    A SubgroupAction looks at subgroups in the regular expression match and
    returns the same amount of tokens as there are subgroups, using the specified
    action for every subgroup.

    For example, the rule::

        "(0x)([0-9a-f]+)", SubgroupAction(Number.Prefix, Number.Hexadecimal)

    yields two tokens in case of a match, one for "0x" and the other for the
    other group of the match.

    There should be the same number of subgroups in the regular expression as
    there are action attributes given to __init__().

    """
    __slots__ = ('_actions',)
    def __init__(self, *actions):
        self._actions = actions

    def replace(self, lexer, pos, text, match):
        for i, action in enumerate(self._actions, match.lastindex + 1):
            yield from lexer.filter_actions(action, match.start(i), match.group(i), match)

    def variations(self):
        """Yield the possible actions."""
        yield from self._actions


class DelegateAction(ActionItem):
    """This action uses a lexicon to parse the text.

    All tokens are yielded as one group, flattened, ignoring the tree
    structure, so this is not efficient for large portions of text, as the
    whole region is parsed again on every modification.

    But it can be useful when you want to match a not too large text blob first
    that's difficult to capture otherwise, and then lex it with a lexicon that
    does (almost) not enter other lexicons.

    """
    __slots__ = ('_lexicon',)

    def __init__(self, lexicon):
        self._lexicon = lexicon

    def replace(self, lexer, pos, text, match):
        """Use our lexicon to parse the matched text."""
        sublexer = type(lexer)([self._lexicon])
        for e in sublexer.events(text):
            for p, txt, action in e.tokens:
                yield pos + p, txt, action

    def evaluate_items(self):
        """Return the lexicon specified on init, used by evaluate() and pre_evaluate()."""
        return self._lexicon,

    def variations(self):
        """Yield our lexicon."""
        yield self._lexicon


class SkipAction(ActionItem):
    """A DynamicAction that yields nothing.

    A SkipAction() is stored in the module variable ``skip`` and causes the rule
    to silently ignore the matched text.

    """
    def replace(self, lexer, pos, text, match):
        yield from ()

    def variations(self):
        """Yield no variations."""
        return
        yield


def evaluate(obj, ns):
    """Evaluate an object, that may or may not be an Item.

    The namespace `ns` is a dictionary containing text, match and/or arg
    variables. A list or a tuple of items is also evaluated and always becomes
    a tuple.

    """
    if isinstance(obj, Item):
        return obj.evaluate(ns)
    if type(obj) in (list, tuple):
        return tuple(evaluate(o, ns) for o in obj)
    return obj


def evaluate_rule(rule, match):
    """Evaluate all RuleItem objects in the rule.

    The specified match object provides the value for the TEXT and MATCH
    variables. Lists and tuples are unrolled.

    """
    ns = {'text': match.group(), 'match': match}
    def eval_rule_items(objs):
        for obj in objs:
            if isinstance(obj, RuleItem):
                yield from unroll(obj.evaluate(ns))
            elif type(obj) in (list, tuple):
                yield from eval_rule_items(obj)
            else:
                yield obj
    yield from eval_rule_items(rule)


def pre_evaluate(obj, ns):
    """Pre-evaluate any object, that may or may not be an Item.

    Returns a two-tuple(result, success). The namespace `ns` is a dictionary
    containing text, match and/or arg variables.

    * If the object is an Item, returns ``object.pre_evaluate(ns)``.
    * If the object is a list or tuple, evaluates the contents and returns a
      tuple.
    * If the object is none of the above, simply returns the object unchanged.

    The ``success`` value can be one of the values described in
    :meth:`Item.pre_evaluate`, or it is 3, meaning that the object is returned
    unchanged and needs no evaluation.

    """
    if isinstance(obj, Item):
        return obj.pre_evaluate(ns)
    if type(obj) in (list, tuple):
        objs, success  = [], 3
        for o in obj:
            res, ok = pre_evaluate(o, ns)
            objs.append(res)
            success &= ok
        return tuple(objs), success
    return obj, 3


def pre_evaluate_rule(rule, arg):
    """Evaluate all RuleItem objects that can be evaluated in the rule.

    The specified ``arg`` provides the value for the ARG variable. Rule items
    that depend on the match object are not yet evaluated. Lists and tuples are
    unrolled. Returns the rule as a tuple.

    """
    ns = {'arg': arg}
    def pre_eval_rule_items(objs):
        for obj in objs:
            if isinstance(obj, RuleItem):
                yield from unroll(obj.pre_evaluate(ns)[0])
            elif type(obj) in (list, tuple):
                yield from pre_eval_rule_items(obj)
            else:
                yield obj
    result = pre_eval_rule_items(rule)
    # the first item may be a pattern instance; it should be evaluated by now
    for item in result:
        if isinstance(item, pattern):
            item = item.value
        return (item,) + tuple(result)
    return ()


def needs_evaluation(rule):
    """Return True if there are items in the rule that need evaluating."""
    for item in rule:
        if isinstance(item, PostponedItem):
            if needs_evaluation(item.evaluate_items()):
                return True
        elif isinstance(item, RuleItem) or \
             (type(item) in (tuple, list) and needs_evaluation(item)):
            return True
    return False


def variations_tree(rule):
    """Return a tuple with the tree structure of all possible variations.

    Branches (choices) are indicated by a frozenset, which contains
    zero or more tuples.

    """
    items = tuple(rule)
    for i, item in enumerate(items):
        if isinstance(item, Item):
            branch = frozenset(variations_tree(unroll(v)) for v in item.variations())
            return (*items[:i], branch, *variations_tree(items[i+1:]))
    else:
        return items


def variations(rule):
    """Yield all possible variations of the rule."""
    items = tuple(rule)
    for i, item in enumerate(items):
        if isinstance(item, Item):
            prefix = items[:i]
            for suffix in variations(items[i+1:]):
                for v in item.variations():
                    for l in variations(unroll(v)):
                        yield prefix + l + suffix
            break
    else:
        yield items


def unroll(obj):
    """Unroll a tuple or list.

    If the object is a tuple or list, yields the unrolled members recursively.
    Otherwise just the object itself is yielded.

    """
    if type(obj) in (tuple, list):
        for i in obj:
            yield from unroll(i)
    else:
        yield obj


#: sentinel denoting that a variation is any integer
a_number = util.Symbol("a_number")

#: sentinel denoting that a variation is any string
a_string = util.Symbol("a_string")




# -*- coding: utf-8 -*-
#
# This file is part of the parce Python package.
#
# Copyright © 2019-2020 by Wilbert Berendsen <info@wilbertberendsen.nl>
#
# This module is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This module is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""
Helper functions and classes for the :mod:`~parce.treebuilder` module.

"""


import collections
import itertools

from parce.lexer import Event, Lexer
from parce.tree import Context, get_group
from parce.target import TargetFactory


#: encapsulates the return values of :meth:`TreeBuilder.build_new_tree`
BuildResult = collections.namedtuple("BuildResult", "tree start end offset lexicons")

#: encapsulates the return values of :meth:`TreeBuilder.replace_tree`
ReplaceResult = collections.namedtuple("ReplaceResult", "start end lexicons")

#: encapsulates the return values of :func:`get_prepared_lexer`
PreparedLexer = collections.namedtuple("PreparedLexer", "lexer events start")


class Changes:
    """Store changes that have to be made to a tree.

    This object is used by
    :meth:`~parce.treebuilder.TreeBuilder.get_changes()`. Calling
    :meth:`add()` merges new changes with the existing changes.

    """
    __slots__ = "text", "root_lexicon", "start", "removed", "added"

    def __init__(self):
        self.text = ""
        self.root_lexicon = False   # meaning no change is requested
        self.start = -1          # meaning no text is altered
        self.removed = 0
        self.added = 0

    def __repr__(self):
        changes = []
        if self.root_lexicon != False:
            changes.append("root_lexicon: {}".format(self.root_lexicon))
        if self.start != -1:
            changes.append("text: {} -{} +{}".format(self.start, self.removed, self.added))
        if not changes:
            changes.append("(no changes)")
        return "<Changes {}>".format(', '.join(changes))

    def add(self, text, root_lexicon=False, start=0, removed=None, added=None):
        """Merge new change with existing changes.

        If added and removed are not given, all text after start is
        considered to be replaced.

        """
        if root_lexicon != False:
            self.root_lexicon = root_lexicon
        if removed is None:
            removed = len(self.text) - start
        if added is None:
            added = len(text) - start
        self.text = text
        if self.start == -1:
            # there were no previous changes
            self.start = start
            self.removed = removed
            self.added = added
            return
        # determine the offset for removed and added
        if start + removed < self.start:
            offset = self.start - start - removed
        elif start > self.start + self.added:
            offset = start - self.start - self.added
        else:
            offset = 0
        # determine which part of removed falls inside existing changes
        start = max(start, self.start)
        end = min(start + removed, self.start + self.added)
        offset -= max(0, end - start)
        # set the new values
        self.start = min(self.start, start)
        self.removed += removed + offset
        self.added += added + offset

    def has_changes(self):
        """Return True when there are actually changes."""
        return self.start != -1 or self.root_lexicon != False

    def new_position(self, pos):
        """Return how the current changes would affect an older start."""
        if pos < self.start:
            return pos
        elif pos < self.start + self.removed:
            return self.start + self.added
        return pos - self.removed + self.added


def get_prepared_lexer(tree, text, start):
    """Get a prepared lexer reading from text, positioned at (or before) start.

    Returns the three-tuple (lexer, events, tokens). The events stream is
    returned seperately because the last Event can be pushed back, so it is
    yielded again. The tokens are the last tokens group that remained the same.

    Returns None when no position to start can be found, just start from the
    beginning in this case.

    """
    while start:
        last_token = start_token = find_token_before(tree, start)
        while last_token and last_token.group is not None:
            group = get_group(last_token)
            if group[-1].end <= start:
                break
            last_token = group[0].previous_token()
        if not last_token:
            return
        # go back at most 10 tokens, to the beginning of a group; if we
        # are at the first token just return 0.
        for start_token in itertools.islice(last_token.backward(), 10):
            pass
        if start_token.group:
            start_token = get_group(start_token)[0]
        start = start_token.pos if start_token.previous_token() else 0
        lexer = get_lexer(start_token) if start else Lexer([tree.lexicon])
        events = lexer.events(text, start)
        # compare the new events with the old tokens; at least one
        # should be the same; if not, go back further if possible
        old_events = events_with_tokens(start_token, last_token)
        prev = None
        for (old, tokens), new in zip(old_events, events):
            if old != new:
                if prev:
                    return lexer, itertools.chain((new,), events), prev
                break
            prev = tokens
        else:
            if prev:
                return lexer, events, prev


def events_with_tokens(start_token, last_token):
    r"""Yield (Event, tokens) tuples for start_token until and including last_token.

    Events are yielded together with token groups (or single tokens in a
    1-length tuple).

    This can be used to compare an existing token structure with events
    originating from a lexer.

    """
    context, start_trail, end_trail = start_token.common_ancestor_with_trail(last_token)
    if context:

        target = TargetFactory()
        get, push, pop = target.get, target.push, target.pop

        def events(nodes):
            stack = []
            i = 0
            n = nodes
            while True:
                z = len(n)
                while i < z:
                    m = n[i]
                    if m.is_context:
                        push(m.lexicon)
                        stack.append(i)
                        i = 0
                        n = m
                        break
                    else:
                        group = m,
                        if m.group is not None:
                            # find the last token in this group
                            for g, j in enumerate(range(i + 1, z), m.group + 1):
                                if n[j].is_context or n[j].group != g:
                                    group = n[i:j]
                                    break
                        tokens = tuple((t.pos, t.text, t.action) for t in group)
                        yield Event(get(), tokens), group
                        i += len(group)
                else:
                    if stack:
                        pop()
                        i = stack.pop() + 1
                        n = n.parent if stack else nodes # slice we started with
                    else:
                        break

        for context, slice_ in context.slices(start_trail, end_trail, target):
            yield from events(context[slice_])


def get_lexer(token):
    """Get a Lexer initialized at the token's ancestry."""
    lexicons = [p.lexicon for p in token.ancestors()]
    lexicons.reverse()
    return Lexer(lexicons)


def new_tree(token):
    """Return an empty context (and its root) with the same ancestry as the token's."""
    c = n = context = Context(token.parent.lexicon, None)
    for p in token.parent.ancestors():
        n = Context(p.lexicon, None)
        c.parent = n
        n.append(c)
        c = n
    return context, n


def find_token_before(node, pos):
    """A version of :meth:`~parce.tree.Context.find_token_before` that can handle
    empty contexts.

    The new tree built inside
    :meth:`~parce.treebuilder.BasicTreeBuilder.build_new` can have an empty
    context at the beginning and/or the end. Returns None if there is no token
    left from pos.

    """
    while True:
        i = 0
        hi = len(node)
        while i < hi:
            mid = (i + hi) // 2
            n = node[mid]
            if n.is_context:
                n = n.first_token() or n    # if no first token, just n itself
            if pos < n.end:
                hi = mid
            else:
                i = mid + 1
        if i == 0:
            return
        node = node[i-1]
        if node.is_token:
            return node

